# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the VariableMapping """

import os
import pytest

from quark import VariableMapping
from quark.io import hdf5


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "variable_mapping.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE


@pytest.fixture
def variable_mappings():
    """ provide the test objects """
    vm_wild = VariableMapping({("s", 0): ("x", 1), ("s", 1): ("x", 2), ("s", 2): ("y", 0, 0)})
    vm_consecutive = VariableMapping({0: ("x", 1), 1: ("x", 2), 2: ("y", 0, 0)})
    yield vm_wild, vm_consecutive, vm_consecutive.inverse

def test_load_hdf5(variable_mappings):
    """ test hdf5 io """
    for i, var_mapping in enumerate(variable_mappings):
        loaded = hdf5.load(VariableMapping, FILENAME_FIX, prefix_group_name=str(i))
        assert loaded == var_mapping

def test_io_hdf5(variable_mappings):
    """ test hdf5 io round trip """
    for var_mapping in variable_mappings:
        hdf5.save(var_mapping, FILENAME_TMP)
        loaded = hdf5.load(VariableMapping, FILENAME_TMP)
        assert loaded == var_mapping
        os.remove(FILENAME_TMP)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" test binary polynomial module """

import pytest
import numpy as np

from quark import Polynomial, PolyBinary, PolyIsing
from quark.poly_binary import get_binary_representation_polynomial, get_unary_representation_polynomial, MEDIUM, \
                              BETTER, STRAIGHT


X, Y, R = "x", "y", "reduction"

P1_DICT = {(1,): 5, (2,): 6, (1, 3): -3, (2, 4): 3, (4,): -1, (): 3}
P2_DICT = {((X, 1), (Y, 2, 3)): -2, ((Y, 1, 2),): 3}
P3_DICT = {(0,): 5, (1, 1): 6, (2, 0): -3, (1, 3): 0.5, (3,): -1, (0, 1): 0.9, (): 16, (2,): 0}
P1_ISING_DICT = {(): 8.0, (1,): 1.75, (1, 3): -0.75, (2,): 3.75, (2, 4): 0.75, (3,): -0.75, (4,): 0.25}
P1_ISING_INV_DICT = {(): 8.0, (1,): -1.75, (1, 3): -0.75, (2,): -3.75, (2, 4): 0.75, (3,): 0.75, (4,): -0.25}

POLY1 = PolyBinary(P1_DICT)
POLY2 = PolyBinary(P2_DICT)
POLY3 = PolyBinary(P3_DICT)
POLY1_NONE = Polynomial(P1_DICT)
POLY1_ISING = PolyIsing(P1_ISING_DICT)
POLY1_ISING_INV = PolyIsing(P1_ISING_INV_DICT, inverted=True)

POLY3_QUADRATIC = [[ 5. ,  0.9, -3. ,  0. ],
                   [ 0. ,  6. ,  0. ,  0.5],
                   [ 0. ,  0. ,  0. ,  0. ],
                   [ 0. ,  0. ,  0. , -1. ]]
POLY3_LINEAR = [ 5.,  6.,  0., -1.]
POLY3_QUAD_NO_DIAG = [[ 0. ,  0.9, -3. ,  0. ],
                      [ 0. ,  0. ,  0. ,  0.5],
                      [ 0. ,  0. ,  0. ,  0. ],
                      [ 0. ,  0. ,  0. ,  0. ]]

QUADRATIC_POLY = PolyBinary({((X, 1),): 7, ((X, 2), (X, 1)): 2, (): 3})
LINEAR_POLY = PolyBinary({((X, 1),): 7, ((X, 2),): 2})

FROM_RIGHT_WARNING = "PolyBinary will be cast to Polynomial when added or multiplied from the right"


def test_init():
    """ test creation """
    poly = PolyBinary({(1, 1, 1): 5, (2,): 9, (2, 2) : -3, (3, 3, 1): -3, (2, 4): 3, (4,): -1, (): 3})
    assert POLY1 == poly
    assert P1_DICT == dict(poly)

    poly = PolyBinary({((Y, 2, 3), (X, 1), (Y, 2, 3), (Y, 2, 3)): -2, ((Y, 1, 2),): 3})
    assert POLY2 == poly
    assert P2_DICT == dict(poly)

def test_repr():
    """ test string representation """
    assert repr(POLY1) == "PolyBinary({(): 3, (1,): 5, (2,): 6, (4,): -1, (1, 3): -3, (2, 4): 3})"

def test_str():
    """ test string representation """
    assert str(POLY1) == "+3 +5 x1 +6 x2 -1 x4 -3 x1 x3 +3 x2 x4"

def test_add():
    """ test addition """
    # pylint: disable=unidiomatic-typecheck
    sum_poly = PolyBinary(P1_DICT)
    sum_poly += Polynomial({(0,) : 1})
    assert type(POLY1 + Polynomial({(0,) : 1})) is PolyBinary
    assert type(sum_poly) is PolyBinary

    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        sum_poly = Polynomial({(4,): 1}) + POLY1
        assert type(sum_poly) is Polynomial

    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        sum_poly = Polynomial({(4,): 1})
        sum_poly += POLY1
        assert type(sum_poly) is Polynomial

    with pytest.raises(ValueError, match="Cannot add PolyBinary and PolyIsing"):
        _ = POLY1 + PolyIsing(P1_DICT)

def test_mul():
    """ test multiplication """
    # pylint: disable=unidiomatic-typecheck
    exp_poly = PolyBinary({(1, 4): 5, (2, 4): 9, (1, 3, 4): -3, (4,): 2})
    prod_poly = POLY1 * PolyBinary({(4,): 1})
    assert prod_poly == exp_poly

    exp_poly = PolyBinary({((X, 1), (Y, 2, 3)): -4, ((X, 1), (Y, 1, 2)): 6})
    prod_poly = POLY2 * PolyBinary({((X, 1),): 2})
    assert prod_poly == exp_poly

    exp_poly = PolyBinary({((X, 1), (Y, 2, 3)): 4, ((Y, 1, 2),): 9, ((X, 1), (Y, 2, 3), (Y, 1, 2)) : -12})
    prod_poly = POLY2 * POLY2
    assert prod_poly == exp_poly

    prod_poly = PolyBinary(P1_DICT)
    prod_poly *= Polynomial({(0,) : 1})
    assert type(POLY1 * Polynomial({(0,) : 1})) is PolyBinary
    assert type(prod_poly) is PolyBinary

    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        prod_poly = Polynomial({(4,): 1}) * POLY1
        assert type(prod_poly) is Polynomial

    with pytest.warns(UserWarning, match=FROM_RIGHT_WARNING):
        prod_poly = Polynomial({(4,): 1})
        prod_poly *= POLY1
        assert type(prod_poly) is Polynomial

    with pytest.raises(ValueError, match="Cannot multiply PolyBinary and PolyIsing"):
        _ = POLY1 * PolyIsing(P1_DICT)

def test_copy():
    """ test copying """
    # pylint: disable=unidiomatic-typecheck
    poly1_copy = POLY1.copy()
    assert poly1_copy == POLY1
    assert dict(poly1_copy) == dict(POLY1)
    assert type(poly1_copy) is PolyBinary

    p2_copy = POLY2.copy()
    assert p2_copy == POLY2
    assert dict(p2_copy) == dict(POLY2)
    assert type(poly1_copy) is PolyBinary

def test_remove_zero_coefficients():
    """ test the removal of monomials with zero coefficient """
    # pylint: disable=unidiomatic-typecheck
    poly_with_zero = {(3,): 0, (2,3): 0}
    poly_with_zero.update(P1_DICT)
    poly = PolyBinary(poly_with_zero)
    assert poly == POLY1
    assert dict(poly) != dict(POLY1)

    poly_nonzero = poly.remove_zero_coefficients()
    assert poly_nonzero == POLY1
    assert dict(poly_nonzero) == dict(POLY1)
    assert type(poly_nonzero) is PolyBinary

def test_preprocess_minimize():
    """ test preprocessing """
    exp_poly = PolyBinary({(): 2})
    exp_variables = {1 : 0, 2 : 0, 3 : 1, 4 : 1}
    preprocessed_poly, preprocessed_variables = POLY1.preprocess_minimize()
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables
    preprocessed_poly, preprocessed_variables = POLY1.preprocess_minimize(unambitious=True)
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables

    exp_poly = PolyBinary({(): -2})
    exp_variables = {(X, 1) : 1, (Y, 2, 3) : 1, (Y, 1, 2) : 0}
    preprocessed_poly, preprocessed_variables = POLY2.preprocess_minimize()
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables
    preprocessed_poly, preprocessed_variables = POLY2.preprocess_minimize(unambitious=True)
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables

    poly = PolyBinary({(0,): 10, (1,): 2, (2,): -8, (3,): 10, (4,): -7, (5,): 4, (0, 1): 2, (0, 2): -10, (0, 3): -10,
                       (1, 3): -8, (1, 4): -7, (2, 3): 8, (2, 5): 1, (3, 5): 2})
    exp_variables = {0: 0, 1: 1, 2: 1, 3: 0, 4: 1, 5: 0}
    preprocessed_poly, preprocessed_variables = poly.preprocess_minimize()
    assert preprocessed_poly == -20
    assert preprocessed_variables == exp_variables

    exp_variables = {1: 1, 2: 1, 4: 1, 5: 0}
    exp_poly = PolyBinary({(): -20, (0,): 2, (3,): 10, (0, 3): -10})
    preprocessed_poly, preprocessed_variables = poly.preprocess_minimize(unambitious=True)
    assert preprocessed_poly == exp_poly
    assert preprocessed_variables == exp_variables

def test_to_ising():
    """ test conversion to Ising polynomials """
    poly_binary = PolyBinary.read_from_string("5 x1 + 6 x2 + 3 x1 x3 - x4 - 3 x2 x4 + 3")
    exp_poly_ising = 0.25 * PolyIsing.read_from_string("32 + 13 x1 + 9 x2 + 3 x3 - 5 x4 + 3 x1 x3 - 3 x2 x4")
    exp_poly_ising_inv = 0.25 * Polynomial.read_from_string("32 - 13 x1 - 9 x2 - 3 x3 + 5 x4 + 3 x1 x3 - 3 x2 x4")
    assert poly_binary.to_ising() == exp_poly_ising
    assert poly_binary.to_ising(inverted=True) == exp_poly_ising_inv

    exp_poly_ising = PolyIsing({(): 8.0, (1,): 1.75, (1, 3): -0.75, (2,): 3.75, (2, 4): 0.75, (3,): -0.75, (4,): 0.25})
    exp_poly_ising_inv = exp_poly_ising.invert()
    assert POLY1.to_ising() == exp_poly_ising
    assert POLY1.to_ising(inverted=True) == exp_poly_ising_inv

    exp_poly_ising = PolyIsing({(): 1.0, ((X, 1),): -0.5, ((Y, 1, 2),): 1.5, ((Y, 2, 3),): -0.5,
                                ((X, 1), (Y, 2, 3)): -0.5})
    exp_poly_ising_inv = exp_poly_ising.invert()
    assert POLY2.to_ising() == exp_poly_ising
    assert POLY2.to_ising(inverted=True) == exp_poly_ising_inv

def test_unknown_poly_to_binary():
    """ test conversion to binary polynomial """
    assert PolyBinary.from_unknown_poly(POLY1) == POLY1

    poly = PolyBinary.from_unknown_poly(POLY1_NONE)
    assert poly == POLY1_NONE
    assert isinstance(poly, PolyBinary)

    assert PolyBinary.from_unknown_poly(POLY1_ISING) == POLY1
    assert PolyBinary.from_unknown_poly(POLY1_ISING_INV) == POLY1

def test_affine_transform():
    """ test affine transformation """
    # pylint: disable=unidiomatic-typecheck
    exp_poly = Polynomial({(4,) : 33, () : 43, (2, 4) : 27, (2,) : 54, (1, 3) : -27, (1,) : -21, (3,) : -36})
    transformed_poly = POLY1.affine_transform(3, 4)
    assert transformed_poly == exp_poly
    assert type(transformed_poly) is Polynomial

def test_evaluate():
    """ test evaluation """
    # pylint: disable=unidiomatic-typecheck
    var_assignment = {(X, 1): -1, (Y, 2, 3): 1}
    evaluated_poly = POLY2.evaluate(var_assignment)
    exp_poly = PolyBinary({(): 2, ((Y, 1, 2),): 3})
    assert evaluated_poly == exp_poly
    assert type(evaluated_poly) is PolyBinary

    var_assignment = {(X, 1) : 0, (Y, 2, 3) : 1, (Y, 1, 2) : 0}
    evaluated_poly = POLY2.evaluate(var_assignment)
    assert evaluated_poly == 0

    with pytest.raises(ValueError, match="Can only assign numbers or polynomials"):
        var_assignment = {(X, 1): PolyBinary({((X, 1),): 2, (): 3})}
        POLY2.evaluate(var_assignment)

    var_assignment = {(X, 1): Polynomial({((X, 1),): 2, (): 3})}
    exp_poly = Polynomial({((X, 1), (Y, 2, 3)): -4, ((Y, 1, 2),): 3, ((Y, 2, 3),): -6})
    evaluated_poly = POLY2.evaluate(var_assignment)
    assert type(evaluated_poly) is Polynomial
    assert evaluated_poly == exp_poly

def test_replace_variables():
    """ test replacement of variables in polynomials """
    # pylint: disable=unidiomatic-typecheck
    replacement = {(X, 1) : "a", (Y, 2, 3) : "b", (Y, 1, 2) : "c"}
    exp_poly = PolyBinary({("a", "b"): -2, ("c",): 3})
    replaced_poly = POLY2.replace_variables(replacement)
    assert replaced_poly == exp_poly
    assert type(replaced_poly) is PolyBinary

def test_get_rounded():
    """ test rounding of polynomials """
    # pylint: disable=unidiomatic-typecheck
    poly = PolyBinary({((X, 1), (Y, 2, 3)): -2.3, ((Y, 1, 2),): 2.9})
    rounded_poly = poly.round(0)
    assert rounded_poly == POLY2
    assert type(rounded_poly) is PolyBinary

def test_get_matrix_representation():
    """ test standard matrix representation """
    with pytest.warns(UserWarning, match="Constant offset of 16 is dropped"):
        quadratic = POLY3.get_matrix_representation()
    assert np.array_equal(quadratic, POLY3_QUADRATIC)

def test_get_from_matrix_representation():
    """ test conversion from matrix representation """
    poly = PolyBinary.get_from_matrix_representation(POLY3_QUADRATIC)
    assert poly == POLY3 - 16

    poly = PolyBinary.get_from_matrix_representation(POLY3_LINEAR, POLY3_QUAD_NO_DIAG)
    assert poly == POLY3 - 16

def test_reduce():
    """ test reduction of polynomial """
    qubic = PolyBinary({(1, 2, 3): 1, (2, 3, 4): 2})
    exp_reduced_poly = PolyBinary({((R, X, 3, R, X, 1, X, 2),): 1,
                                   ((R, X, 4, R, X, 2, X, 3),): 2})
    exp_reductions = [((R, X, 1, X, 2), (X, 1), (X, 2)),
                      ((R, X, 2, X, 3), (X, 2), (X, 3)),
                      ((R, X, 3, R, X, 1, X, 2), (X, 3), (R, X, 1, X, 2)),
                      ((R, X, 4, R, X, 2, X, 3), (X, 4), (R, X, 2, X, 3))]

    with pytest.warns(UserWarning, match="Variables are replaced by tuple variables"):
        reduced_poly, reductions = qubic.reduce(reduction_variable_prefix=R)
    assert reduced_poly.degree == 1
    assert reduced_poly == exp_reduced_poly
    assert reductions == exp_reductions

    exp_reduced_poly = PolyBinary({((R, X, 1, X, 2), (X, 3)): 1,
                                   ((R, X, 2, X, 3), (X, 4)): 2})
    exp_reductions = [((R, X, 1, X, 2), (X, 1), (X, 2)),
                      ((R, X, 2, X, 3), (X, 2), (X, 3))]

    with pytest.warns(UserWarning, match="Variables are replaced by tuple variables"):
        reduced_poly, reductions = qubic.reduce(max_degree=2)
    assert reduced_poly.degree == 2
    assert reduced_poly == exp_reduced_poly
    assert reductions == exp_reductions

    use = {((R, X, 2, X, 3), (X, 2), (X, 3))}
    exp_reduced_poly = PolyBinary({((R, X, 2, X, 3), (X, 1)): 1,
                                   ((R, X, 2, X, 3), (X, 4)): 2})
    with pytest.warns(UserWarning, match="Variables are replaced by tuple variables"):
        reduced_poly, reductions = qubic.reduce(max_degree=2, use=use)
    assert reduced_poly.degree == 2
    assert reduced_poly == exp_reduced_poly
    assert not reductions

    use = [(10, 2, 3)]
    exp_reduced_poly = PolyBinary({(10, 1): 1, (10, 4): 2})
    reduced_poly, reductions = qubic.reduce(max_degree=2, use=use)
    assert reduced_poly.degree == 2
    assert reduced_poly == exp_reduced_poly
    assert not reductions

    reduced_poly, reductions = qubic.reduce(max_degree=3)
    assert reduced_poly.degree == 3
    assert reduced_poly == qubic
    assert not reductions

    with pytest.raises(ValueError, match="Unknown strategy 'sth_else'"):
        with pytest.warns(UserWarning, match="Variables are replaced by tuple variables"):
            qubic.reduce(reduction_strategy="sth_else")

    with pytest.raises(ValueError, match="Provide integer max_degree larger than 0 or set to None"):
        qubic.reduce(max_degree=0)

def test_reduce_options():
    """ test reduction of polynomial with different options to choose the next variable pair to be replaced """
    poly = PolyBinary({(1, 2, 3, 4): 1, (2, 3, 4, 5): 2, (3, 4): 3})

    exp_poly3 = PolyBinary({((X, 3), (X, 4)): 3,
                            ((R, X, 1, X, 2), (X, 3), (X, 4)): 1,
                            ((R, X, 2, X, 3), (X, 4), (X, 5)): 2})
    exp_reductions3 = [((R, X, 1, X, 2), (X, 1), (X, 2)),
                       ((R, X, 2, X, 3), (X, 2), (X, 3))]
    exp_poly2 = PolyBinary({((R, X, 3, X, 4),): 3,
                            ((R, X, 1, X, 2), (R, X, 3, X, 4)): 1,
                            ((R, X, 2, X, 5), (R, X, 3, X, 4)): 2})
    exp_reductions2 = [((R, X, 1, X, 2), (X, 1), (X, 2)),
                       ((R, X, 2, X, 5), (X, 2), (X, 5)),
                       ((R, X, 3, X, 4), (X, 3), (X, 4))]

    _check_reductions(STRAIGHT, exp_poly2, exp_poly3, exp_reductions2, exp_reductions3, poly)

    exp_poly3 = PolyBinary({((R, X, 2, X, 3), (X, 1), (X, 4)): 1,
                            ((R, X, 2, X, 3), (X, 4), (X, 5)): 2,
                            ((X, 3), (X, 4)): 3})
    exp_reductions3 = [((R, X, 2, X, 3), (X, 2), (X, 3))]
    exp_poly2 = PolyBinary({((R, X, 4, R, X, 2, X, 3), (X, 1)): 1,
                            ((R, X, 4, R, X, 2, X, 3), (X, 5)): 2,
                            ((X, 3), (X, 4)): 3})
    exp_reductions2 = [((R, X, 2, X, 3), (X, 2), (X, 3)),
                       ((R, X, 4, R, X, 2, X, 3), (X, 4), (R, X, 2, X, 3))]

    _check_reductions(MEDIUM, exp_poly2, exp_poly3, exp_reductions2, exp_reductions3, poly)

    exp_poly3 = PolyBinary({((R, X, 3, X, 4),): 3,
                             ((R, X, 3, X, 4), (X, 1), (X, 2)): 1,
                             ((R, X, 3, X, 4), (X, 2), (X, 5)): 2})
    exp_reductions3 = [((R, X, 3, X, 4), (X, 3), (X, 4))]
    exp_poly2 = PolyBinary({((R, X, 3, X, 4),): 3,
                            ((R, X, 2, R, X, 3, X, 4), (X, 1)): 1,
                            ((R, X, 2, R, X, 3, X, 4), (X, 5)): 2})
    exp_reductions2 = [((R, X, 2, R, X, 3, X, 4), (X, 2), (R, X, 3, X, 4)),
                       ((R, X, 3, X, 4), (X, 3), (X, 4))]

    _check_reductions(BETTER, exp_poly2, exp_poly3, exp_reductions2, exp_reductions3, poly)

def _check_reductions(reduction_strategy, exp_poly2, exp_poly3, exp_reductions2, exp_reductions3, poly):
    with pytest.warns(UserWarning, match="Variables are replaced by tuple variables"):
        reduced_poly3, reductions3 = poly.reduce(max_degree=3, reduction_strategy=reduction_strategy)
        reduced_poly2, reductions2 = poly.reduce(max_degree=2, reduction_strategy=reduction_strategy)
    assert reduced_poly3.degree == 3
    assert reduced_poly3 == exp_poly3
    assert reductions3 == exp_reductions3
    assert reduced_poly2.degree == 2
    assert reduced_poly2 == exp_poly2
    assert reductions2 == exp_reductions2


def test_reduce_quadratic_to_linear_polynomial():
    """ test reduction to quadratic polynomial """
    exp_poly = PolyBinary({((R, X, 1, X, 2),): 2, ((X, 1),): 7, (): 3})
    exp_reductions = [((R, X, 1, X, 2), (X, 1), (X, 2))]
    reduced_poly, reductions = QUADRATIC_POLY.reduce()
    assert reduced_poly == exp_poly
    assert reductions == exp_reductions

    flat_poly = PolyBinary({(2, 1): 2, (1,): 7, (): 3})
    with pytest.warns(UserWarning, match="Variables are replaced by tuple variables"):
        reduced_poly, reductions = flat_poly.reduce()
    assert reduced_poly == exp_poly
    assert reductions == exp_reductions

    reduced_poly, reductions = LINEAR_POLY.reduce()
    assert reduced_poly == LINEAR_POLY
    assert not reductions

def test_get_binary_representation_polynomial():
    """ test construction of binary representation polynomial """
    poly = get_binary_representation_polynomial(0)
    exp_poly = PolyBinary()
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(1)
    exp_poly = PolyBinary({((X, 0),): 1})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(2)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 1})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(3)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 2})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(4)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 2, ((X, 2),): 1})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(5, 1)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 2, ((X, 2),): 1, (): 1})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(5, -1)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 2, ((X, 2),): 3, (): -1})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(5, -2, split_signs=True)
    exp_poly = PolyBinary({((X, "+", 0),): 1, ((X, "+", 1),): 2, ((X, "+", 2),): 2,
                           ((X, "-", 0),): -1, ((X, "-", 1),): -1})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(-2, -7)
    exp_poly = PolyBinary({((X, 0),): -1, ((X, 1),): -2, ((X, 2),): -2, (): -2})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(num_variables=3)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 2, ((X, 2),): 4})
    assert poly == exp_poly

    poly = get_binary_representation_polynomial(7, variable_prefix="slack")
    exp_poly = PolyBinary({(("slack", 0),): 1, (("slack", 1),): 2, (("slack", 2),): 4})
    assert poly == exp_poly

    with pytest.raises(ValueError, match="Either provide upper_bound or num_variables"):
        get_binary_representation_polynomial()

    with pytest.warns(UserWarning, match="With upper_bound given, num_variables is ignored"):
        get_binary_representation_polynomial(7, 3, 3)

    with pytest.raises(ValueError, match="Upper bound needs to be larger than or equal to lower bound"):
        get_binary_representation_polynomial(-7)

def test_get_unary_representation_polynomial():
    """ test construction of unary representation polynomial """
    poly = get_unary_representation_polynomial(0)
    exp_poly = PolyBinary()
    assert poly == exp_poly

    poly = get_unary_representation_polynomial(1)
    exp_poly = PolyBinary({((X, 0),): 1})
    assert poly == exp_poly

    poly = get_unary_representation_polynomial(2)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 1})
    assert poly == exp_poly

    poly = get_unary_representation_polynomial(3)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 1, ((X, 2),): 1})
    assert poly == exp_poly

    poly = get_unary_representation_polynomial(4, 1)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 1, ((X, 2),): 1, (): 1})
    assert poly == exp_poly

    poly = get_unary_representation_polynomial(3, -1)
    exp_poly = PolyBinary({((X, 0),): 1, ((X, 1),): 1, ((X, 2),): 1, ((X, 3),): 1, (): -1})
    assert poly == exp_poly

    poly = get_unary_representation_polynomial(-2, -5)
    exp_poly = PolyBinary({((X, 0),): -1, ((X, 1),): -1, ((X, 2),): -1, (): -2})
    assert poly == exp_poly

    poly = get_unary_representation_polynomial(3, variable_prefix="slack")
    exp_poly = PolyBinary({(("slack", 0),): 1, (("slack", 1),): 1, (("slack", 2),): 1})
    assert poly == exp_poly

    with pytest.raises(ValueError, match="Upper bound needs to be larger than or equal to lower bound"):
        get_unary_representation_polynomial(-7)

def test_positive_negative():
    """ test positive and negative part """
    exp_positive = PolyBinary({(1,): 5, (2,): 6, (2, 4): 3})
    exp_negative = PolyBinary({(1, 3): -3, (4,): -1})
    assert POLY1.positive == exp_positive
    assert POLY1.negative == exp_negative

def test_naive_upper_bound():
    """ test upper bound """
    assert POLY1.naive_upper_bound == 17
    assert POLY2.naive_upper_bound == 3

def test_naive_lower_bound():
    """ test lower bound"""
    assert POLY1.naive_lower_bound == -1
    assert POLY2.naive_lower_bound == -2



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the logging helper functions """

import os
import shutil
import logging
from pathlib import Path

from quark.utils import get_logger


TEST_DIR = os.path.abspath(os.path.curdir) + "/"


def test_defaults():
    """ test default initialization """
    logger_default = get_logger(logger_name="logger_default")
    assert logger_default.name == "logger_default"
    assert logger_default.level == logging.INFO
    handlers = logger_default.handlers
    assert len(handlers) == 1
    handler = handlers[0]
    assert isinstance(handler, logging.StreamHandler)
    assert handler.level == logging.INFO

def test_empty_logger():
    """ test empty logger """
    logger_empty = get_logger(logger_name="logger_empty",
                              log_level_console=logging.NOTSET)
    assert logger_empty.handlers == []

def test_threshold_log_level():
    """ test if threshold loglevel overwrites handler loglevels"""
    logger_threshold_loglevel = get_logger(logger_name="logger_threshold_loglevel",
                                           log_level=logging.WARNING)
    assert logger_threshold_loglevel.level == logging.WARNING

def test_file_name_default():
    """ test default values if file logging is enables"""
    logger_file_default = get_logger(logger_name="logger_file_default",
                                     log_level=logging.INFO,
                                     log_level_console=logging.ERROR,
                                     log_level_file=logging.INFO)
    handlers = logger_file_default.handlers
    assert len(handlers) == 2
    console_handler = handlers[0]
    file_handler = handlers[1]
    assert isinstance(console_handler, logging.StreamHandler)
    assert isinstance(file_handler, logging.FileHandler)
    assert console_handler.level == logging.ERROR
    assert file_handler.level == logging.INFO
    logfile = TEST_DIR + "logger_file_default.log"
    assert Path(file_handler.baseFilename) == Path(logfile)
    
    tear_down(logger_file_default, logfile)

def test_file_name_explicit():
    """ test explicit naming of logfile"""
    logger_file_explicit = get_logger(logger_name="logger_file_explicit",
                                      log_level=logging.INFO,
                                      log_level_console=logging.ERROR,
                                      log_level_file=logging.INFO,
                                      log_file="mylogfile.log")
    handlers = logger_file_explicit.handlers
    assert len(handlers) == 2
    file_handler = handlers[1]
    logfile = TEST_DIR + "mylogfile.log"
    assert isinstance(file_handler, logging.FileHandler)
    assert Path(file_handler.baseFilename) == Path(logfile)

    tear_down(logger_file_explicit, logfile)

def test_folder_and_file_name_explicit():
    """ test explicit naming of logfile and logfolder"""
    logger_folder_and_file_explicit = get_logger(logger_name="logger_folder_and_file_explicit",
                                                 log_level=logging.INFO,
                                                 log_level_console=logging.ERROR,
                                                 log_level_file=logging.INFO,
                                                 log_file="mylogfile.log",
                                                 log_folder=TEST_DIR + "mylogfolder/")
    handlers = logger_folder_and_file_explicit.handlers
    assert len(handlers) == 2
    file_handler = handlers[1]
    logfolder = TEST_DIR + "mylogfolder/"
    logfile = logfolder + "mylogfile.log"
    assert isinstance(file_handler, logging.FileHandler)
    assert Path(file_handler.baseFilename) == Path(logfile)

    tear_down(logger_folder_and_file_explicit, logfile, logfolder)

def test_folder_name_explicit():
    """ test explicit naming of logfolder but default nameing of logfile"""
    logger_folder_explicit = get_logger(logger_name="logger_folder_explicit",
                                        log_level=logging.INFO,
                                        log_level_console=logging.ERROR,
                                        log_level_file=logging.INFO,
                                        log_folder=TEST_DIR + "mylogfolder/")
    handlers = logger_folder_explicit.handlers
    assert len(handlers) == 2
    file_handler = handlers[1]
    logfolder = TEST_DIR + "mylogfolder/"
    logfile = logfolder + "logger_folder_explicit.log"
    assert isinstance(file_handler, logging.FileHandler)
    assert Path(file_handler.baseFilename) == Path(logfile)

    tear_down(logger_folder_explicit, logfile, logfolder)


def tear_down(logger, logfile, logfolder=None):
    """ closing handlers and removing file resp. folder """
    logger.close_all_handlers()
    assert os.path.exists(logfile)
    os.remove(logfile)
    if logfolder:
        assert os.path.exists(logfolder)
        shutil.rmtree(logfolder)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for helper functions for storing different types of Polynomials """

from quark import Polynomial, PolyIsing, PolyBinary
from . import hdf5, hdf5_attributes


# Register mapping from group name to type
POLY_TYPES = [Polynomial, PolyIsing, PolyBinary]
POLY_STR_2_TYPE = {hdf5.get_group_name(poly_type): poly_type for poly_type in POLY_TYPES}

TYPE = "type"

ERROR_GROUP    = "Did not find Polynomial '{}' in group '{}' in hdf5 file '{}'"
ERROR_TOO_MUCH = "Found too much Polynomials in group '{}'"
ERROR_NO       = "Found no Polynomials in group '{}'"


def write_poly_with_type(group, polynomial, name=None):
    """
    write the polynomial in the opened group of an HDF5 file,
    either under the provided name or under the default group name for the Polynomial type
    (thus only one polynomial can be stored in the group if used without name)

    :param (h5py.Group) group: the group to store the data in
    :param (Polynomial) polynomial: the polynomial that shall be stored
    :param (str or None) name: the name of the group under which the polynomial is stored,
                               by default the Polynomial type is used
    """
    group_name = name or hdf5.get_group_name(type(polynomial))
    hdf5.save_in(polynomial, group, group_name=group_name)
    if name:
        # if the polynomial is stored under a different name, save the type as an attribute
        _add_poly_type(group, polynomial, name)

def _add_poly_type(group, polynomial, name):
    # if a name is given, the type of the Polynomial needs to be stored additionally
    poly_group = group[name]
    poly_type_str = hdf5.get_group_name(type(polynomial))
    # the type is added as an attribute
    hdf5_attributes.write_attribute(poly_group, TYPE, value=poly_type_str)


def read_poly_with_type(group, name=None):
    """
    read the polynomial from the opened group of an HDF5 file,
    if no name is given, the default group name is the Polynomial type
    (thus there should only be one Polynomial in the group without a name)

    :param (h5py.Group) group: the group to read the data from
    :param (str or None) name: the name of the group under which the polynomial is stored,
                               by default the first found Polynomial type is used
    :return: the loaded Polynomial
    """
    poly_type = _get_poly_type(group, name)
    group_name = name or hdf5.get_group_name(poly_type)
    return hdf5.load_from(poly_type, group, group_name=group_name)

def _get_poly_type(group, name=None):
    if name:
        if name not in group:
            raise ValueError(ERROR_GROUP.format(name, group.name, group.file.filename))
        poly_group = group[name]
        # read the type of the polynomial
        poly_type_str = hdf5_attributes.read_attribute(poly_group, TYPE, check_existence=False)
        # if we found no type check if the name itself corresponds to a valid type
        if not poly_type_str and name in POLY_STR_2_TYPE:
            poly_type_str = name
    else:
        # if no name is given, we need to find the stored Polynomial
        # check if there is some Polynomial among the stored objects in the group
        poly_type_strs = [entry for entry in group.keys() if entry in POLY_STR_2_TYPE]
        if len(poly_type_strs) > 1:
            # we do not know which Polynomial to choose when there are several
            raise ValueError(ERROR_TOO_MUCH.format(group.name))
        if len(poly_type_strs) < 1:
            raise ValueError(ERROR_NO.format(group.name))
        poly_type_str = poly_type_strs[0]

    poly_type = POLY_STR_2_TYPE.get(poly_type_str, Polynomial)
    return poly_type



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO with HDF5 files """

import os
import pytest
import numpy as np
import h5py

from quark.io import hdf5, hdf5_datasets


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "example_object.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE

NAME = "example_object"
VALUE = 99
DATA = {"z" : 1.0, "y" : 2.5, "x" : 4.0}

PREFIX = "Prefix"
EXAMPLE_OBJECT = "ExampleObject"
HALF_GROUP_NAME = "ExampleObject/example_object"
FULL_GROUP_NAME_99 = "ExampleObject/example_object/99"


class ExampleObject:
    """ an object with an io implementation """

    def __init__(self, data=None, name=None, value=None):
        self.data = data
        self.name = name
        self.value = value

    @staticmethod
    def get_identifying_attributes():
        """ identifying attributes """
        return ["name", "value"]

    def write_hdf5(self, group):
        """ necessary write method for io """
        hdf5_datasets.write_dataset(group, "data", self)

    @staticmethod
    def read_hdf5(group):
        """ necessary read method for io """
        data = hdf5_datasets.read_dataset(group, "data")
        return dict(data=data)

class ExampleObjectWrongIO:
    """ an object with an io implementation """

    def __init__(self, name, data, value=None):
        self.data = data
        self.name = name
        self.value = value

    @staticmethod
    def get_identifying_attributes():
        """ identifying attributes """
        return ["name", "value"]

    @staticmethod
    def read_hdf5(group):
        """ necessary read method for io """
        return dict()

class ExampleObjectWithoutIO:
    """ a different object without an io implementation """

    def __init__(self):
        pass

class TwoLevelExampleObject:
    """ a new object with an io implementation but calling ExampleObjectWithoutIO's non-existing io directly """

    def __init__(self):
        self.obj = ExampleObjectWithoutIO()

    def write_hdf5(self, group):
        """ necessary write method for io """
        hdf5.save_in(self.obj, group)

    @staticmethod
    def read_hdf5(group):
        """ necessary read method for io """
        return hdf5.load_data_from(ExampleObjectWithoutIO, group)

class InheritingExampleObject(ExampleObjectWithoutIO):
    """ a new object with an io implementation but calling ExampleObjectWithoutIO's non-existing io by inheritance """

    def write_hdf5(self, group):
        """ necessary write method for io """
        hdf5.save_in(super(), group)

    @staticmethod
    def read_hdf5(group):
        """ necessary read method for io """
        return hdf5.load_data_from(super(InheritingExampleObject, InheritingExampleObject), group)

@pytest.fixture
def example_object():
    """ provide test object of example class """
    yield ExampleObject(DATA, NAME, VALUE)


def test_exists():
    """ test exist method """
    # check for full identifying attributes - ExampleObject/example_object/99 exists
    assert hdf5.exists(ExampleObject, FILENAME_FIX, name=NAME, value=VALUE)

    # check for only name as identifying attributes - ExampleObject/example_object exists
    assert hdf5.exists(ExampleObject, FILENAME_FIX, name=NAME)

    # check for only value as identifying attributes - ExampleObject/99 does not exist (name inbetween missing)
    assert not hdf5.exists(ExampleObject, FILENAME_FIX, value=VALUE)

    # check for some group name - ExampleObject/example_object exists
    assert hdf5.exists(ExampleObject, FILENAME_FIX, full_group_name=HALF_GROUP_NAME)

    # check for full group name - ExampleObject/example_object/99 exists
    assert hdf5.exists(ExampleObject, FILENAME_FIX, full_group_name=FULL_GROUP_NAME_99)

    # check for full group name - ExampleObject/example_object/100 does not exist
    assert not hdf5.exists(ExampleObject, FILENAME_FIX, full_group_name=HALF_GROUP_NAME + "/100")

    # check for file - does not exist
    assert not hdf5.exists(ExampleObject, "testdata/no_object.h5", name=NAME)

    # check for prefixed versions - do exist
    assert hdf5.exists(ExampleObject, FILENAME_FIX, prefix_group_name=PREFIX, name=NAME, value=VALUE)
    assert hdf5.exists(ExampleObject, FILENAME_FIX, full_group_name=PREFIX + "/" + FULL_GROUP_NAME_99)

    # check for option clash
    with pytest.raises(ValueError, match="Either provide prefix for default group name or full group name"):
        assert hdf5.exists(ExampleObject, FILENAME_FIX, prefix_group_name=PREFIX, full_group_name=FULL_GROUP_NAME_99)

def test_save(example_object):
    """ test if the object is saved correctly, where we will use here ExampleObject exemplary """
    hdf5.save(example_object, FILENAME_TMP)
    # check if the expected groups are in the file
    with h5py.File(FILENAME_TMP, "r") as h5file:
        subgroup = _check_structure_get_data_group(h5file)
        assert "keys" in subgroup
        assert "values" in subgroup
        assert np.array_equal(subgroup["keys"][()], [b"z", b"y", b"x"])
        assert np.array_equal(subgroup["values"][()], [1.0, 2.5, 4.0])

    # creating new ExampleObject with different Polynomial but same name
    new_data = {"c": 2, "a": 3, "b": 4}
    new_obj = ExampleObject(new_data, NAME, VALUE)

    message = f"Group 'ExampleObject/example_object/99' already existed in file '.*{FILE}' and got deleted"
    with pytest.warns(UserWarning, match=message):
        hdf5.save(new_obj, FILENAME_TMP, mode="a")
    # should have overwritten the old ExampleObject in the file
    with h5py.File(FILENAME_TMP, "r") as h5file:
        subgroup = _check_structure_get_data_group(h5file)
        assert np.array_equal(subgroup["keys"][()], [b"c", b"a", b"b"])
        assert np.array_equal(subgroup["values"][()], [2, 3, 4])

    # creating new ExampleObject with different value
    new_value = 100
    new_obj = ExampleObject(DATA, NAME, new_value)
    hdf5.save(new_obj, FILENAME_TMP)
    # now there should be two ExampleObjects in the file
    with h5py.File(FILENAME_TMP, "r") as h5file:
        assert EXAMPLE_OBJECT in h5file
        group = h5file[EXAMPLE_OBJECT][NAME]
        assert str(VALUE) in group
        assert str(new_value) in group

    hdf5.save(new_obj, FILENAME_TMP, full_group_name="Elsewhere/different", mode="a")
    with h5py.File(FILENAME_TMP, "r") as h5file:
        assert "Elsewhere" in h5file
        group = h5file["Elsewhere"]["different"]
        assert "data" in group

    with pytest.raises(ValueError, match="Either provide prefix for default group name or full group name"):
        assert hdf5.save(ExampleObject, FILENAME_TMP, prefix_group_name=PREFIX, full_group_name=FULL_GROUP_NAME_99)

    with pytest.raises(ValueError, match="Object of type 'ExampleObjectWrongIO' does not have a write_hdf5 method"):
        hdf5.save(ExampleObjectWrongIO("name", 1), FILENAME_TMP)

    os.remove(FILENAME_TMP)

def _check_structure_get_data_group(h5file):
    assert EXAMPLE_OBJECT in h5file
    group = h5file[EXAMPLE_OBJECT]
    assert NAME in group
    subgroup = group[NAME]
    assert str(VALUE) in subgroup
    subgroup = subgroup[str(VALUE)]
    assert "data" in subgroup
    subgroup = subgroup["data"]
    return subgroup

def test_save_without_all_identifying_attributes():
    """
    testing if the object is saved correctly
    we will use here ExampleObject exemplary
    """
    # with name but without value
    example_object = ExampleObject(DATA, name=NAME)
    hdf5.save(example_object, FILENAME_TMP)
    # check if the expected groups are in the file
    with h5py.File(FILENAME_TMP, "r") as h5file:
        assert EXAMPLE_OBJECT in h5file
        group = h5file[EXAMPLE_OBJECT]
        assert NAME in group
        subgroup = group[NAME]
        assert not str(VALUE) in subgroup
        assert "data" in subgroup
    os.remove(FILENAME_TMP)

    # with value but without name
    example_object = ExampleObject(DATA, value=VALUE)
    with pytest.warns(UserWarning, match="Intermediate identifier is None, group name cannot be parsed back"):
        hdf5.save(example_object, FILENAME_TMP)
    # check if the expected groups are in the file
    with h5py.File(FILENAME_TMP, "r") as h5file:
        assert EXAMPLE_OBJECT in h5file
        group = h5file[EXAMPLE_OBJECT]
        assert not NAME in group
        assert str(VALUE) in group
        subgroup = group[str(VALUE)]
        assert "data" in subgroup
    os.remove(FILENAME_TMP)

    # without name and value
    example_object = ExampleObject(DATA)
    hdf5.save(example_object, FILENAME_TMP)
    # check if the expected groups are in the file
    with h5py.File(FILENAME_TMP, "r") as h5file:
        assert EXAMPLE_OBJECT in h5file
        group = h5file[EXAMPLE_OBJECT]
        assert not NAME in group
        assert not str(VALUE) in group
        assert "data" in group
    os.remove(FILENAME_TMP)

def test_load(example_object):
    """
    testing if the object is loaded correctly
    we will use here ExampleObject exemplary
    """
    # load with all identifying attributes - ExampleObject/example_object/99
    loaded = hdf5.load(ExampleObject, FILENAME_FIX, name=NAME, value=VALUE)
    _check_loaded(loaded, example_object)

    # load with all identifying attributes and prefix - Prefix/ExampleObject/example_object/99
    loaded = hdf5.load(ExampleObject, FILENAME_FIX, prefix_group_name=PREFIX, name=NAME, value=VALUE)
    _check_loaded(loaded, example_object)

    # load with full group name containing all identifying attributes - ExampleObject/example_object/99
    with pytest.warns(UserWarning, match="Experimental parsing of attributes"):
        loaded = hdf5.load(ExampleObject, FILENAME_FIX, full_group_name=FULL_GROUP_NAME_99)
    _check_loaded(loaded, example_object)

    # missing identifying attributes
    with pytest.raises(ValueError, match=f"Did not find dataset 'data' in group '/{EXAMPLE_OBJECT}' in HDF5 file"):
        hdf5.load(ExampleObject, FILENAME_FIX)
    with pytest.raises(ValueError, match=f"Did not find dataset 'data' in group '/{HALF_GROUP_NAME}' in HDF5 file"):
        hdf5.load(ExampleObject, FILENAME_FIX, name=NAME)

    wrong_name = EXAMPLE_OBJECT + '/wrong_name'
    with pytest.raises(ValueError, match=f"Did not find group '{wrong_name}' in HDF5 file '.*{FILE}'"):
        hdf5.load(ExampleObject, FILENAME_FIX, name='wrong_name')
    with pytest.warns(UserWarning, match='Experimental parsing of attributes'):
        with pytest.raises(ValueError, match=f"Did not find group '{wrong_name}' in HDF5 file '.*{FILE}'"):
            hdf5.load(ExampleObject, FILENAME_FIX, full_group_name=wrong_name)
    with pytest.raises(ValueError, match='Either provide prefix for default group name or full group name'):
        assert hdf5.load(ExampleObject, FILENAME_FIX, prefix_group_name=PREFIX, full_group_name=FULL_GROUP_NAME_99)
    with pytest.raises(AssertionError, match="IO is not implemented correctly"):
        with pytest.warns(UserWarning, match='Experimental parsing of attributes'):
            hdf5.load(ExampleObjectWrongIO, FILENAME_FIX, full_group_name=FULL_GROUP_NAME_99)
    with pytest.raises(ValueError, match="Either give identifier 'name' of the class or a full group name"):
        hdf5.load(ExampleObjectWrongIO, FILENAME_FIX, value="example_object/99")

def test_load_identifiers(example_object):
    """ test save and load round trip with different identifier settings """
    # with all identifiers
    hdf5.save(example_object, filename=FILENAME_TMP)
    loaded = hdf5.load(ExampleObject, FILENAME_TMP, name=NAME, value=VALUE)
    _check_loaded(loaded, example_object)
    with pytest.warns(UserWarning, match="Experimental parsing of attributes from group name"):
        loaded = hdf5.load(ExampleObject, FILENAME_TMP, full_group_name=FULL_GROUP_NAME_99)
    _check_loaded(loaded, example_object)
    os.remove(FILENAME_TMP)

    # with name but without value
    example_object = ExampleObject(DATA, NAME, None)
    hdf5.save(example_object, filename=FILENAME_TMP)
    loaded = hdf5.load(ExampleObject, FILENAME_TMP, name=NAME)
    _check_loaded(loaded, example_object)
    loaded = hdf5.load(ExampleObject, FILENAME_TMP, name=NAME, value=None)
    _check_loaded(loaded, example_object)
    with pytest.warns(UserWarning, match="Experimental parsing of attributes from group name"):
        loaded = hdf5.load(ExampleObject, FILENAME_TMP, full_group_name=HALF_GROUP_NAME)
    _check_loaded(loaded, example_object)
    os.remove(FILENAME_TMP)

    # without name and value
    example_object = ExampleObject(DATA, None, None)
    hdf5.save(example_object, filename=FILENAME_TMP)
    loaded = hdf5.load(ExampleObject, FILENAME_TMP)
    _check_loaded(loaded, example_object)
    loaded = hdf5.load(ExampleObject, FILENAME_TMP, name=None)
    _check_loaded(loaded, example_object)
    os.remove(FILENAME_TMP)

    # with value but without name
    example_object = ExampleObject(DATA, None, VALUE)
    with pytest.warns(UserWarning, match="Intermediate identifier is None, group name cannot be parsed back"):
        hdf5.save(example_object, filename=FILENAME_TMP)
    loaded = hdf5.load(ExampleObject, FILENAME_TMP, name=None, value=VALUE)
    _check_loaded(loaded, example_object)
    loaded = hdf5.load(ExampleObject, FILENAME_TMP, value=VALUE)
    _check_loaded(loaded, example_object)

    with pytest.warns(UserWarning, match="Experimental parsing of attributes from group name"):
        loaded = hdf5.load(ExampleObject, FILENAME_TMP, full_group_name="ExampleObject/99")
    assert loaded.data == example_object.data
    # here something went wrong because it cannot be retrieved which identifier was set to None
    # as the ordering of the identifiers is important!
    assert loaded.name == example_object.value
    assert loaded.value is None
    os.remove(FILENAME_TMP)

def _check_loaded(loaded, saved):
    assert loaded.name == saved.name
    assert loaded.value == saved.value
    assert loaded.data == saved.data

def test_error():
    """ testing non-existing io on two test objects """
    _test_error_on(ExampleObjectWithoutIO)
    _test_error_on(TwoLevelExampleObject)

def _test_error_on(cls):
    example_object = cls()
    # trying to store the object is failing
    with pytest.raises(ValueError, match="Object of type 'ExampleObjectWithoutIO' does not have a write_hdf5 method"):
        hdf5.save(example_object, FILENAME_TMP)
    with pytest.raises(ValueError, match="Class 'ExampleObjectWithoutIO' does not have a read_hdf5 method"):
        hdf5.load(cls, FILENAME_TMP)
    os.remove(FILENAME_TMP)

def test_inheritance_error():
    """ testing non-existing io on inheriting object """
    example_object = InheritingExampleObject()
    # trying to store the object is failing
    with pytest.raises(ValueError, match="Object of type 'super' does not have a write_hdf5 method"):
        hdf5.save(example_object, FILENAME_TMP)
    with pytest.raises(ValueError, match=r"\(Super\?\) Class does not have a read_hdf5 method"):
        hdf5.load(InheritingExampleObject, FILENAME_TMP)
    os.remove(FILENAME_TMP)

def test_is_subgroup():
    """ test helper method for subgroups """
    with h5py.File(FILENAME_FIX, "r") as h5file:
        assert hdf5.is_subgroup(h5file, EXAMPLE_OBJECT)
        group = h5file[EXAMPLE_OBJECT]
        assert hdf5.is_subgroup(group, NAME)
        assert not hdf5.is_subgroup(group, "wrong_name")

def test_parse_identifiers_from_group_name():
    """ test helper method for identifiers """
    exp_identifiers = {"name" : NAME, "value" : 100}
    with pytest.warns(UserWarning, match="Experimental parsing of attributes from group name"):
        identifiers = hdf5.parse_identifiers_from_group_name(ExampleObject, HALF_GROUP_NAME + "/100")
    assert identifiers == exp_identifiers

def test_get_all_subgroup_names():
    """ test helper method for getting all subgroups """
    names = hdf5.get_all_subgroup_names(FILENAME_FIX)
    assert names == [EXAMPLE_OBJECT, "ExampleObjectWrongIO", PREFIX]

    names = hdf5.get_all_subgroup_names(FILENAME_FIX, EXAMPLE_OBJECT)
    assert names == [HALF_GROUP_NAME]

    names = hdf5.get_all_subgroup_names(FILENAME_FIX, EXAMPLE_OBJECT, depth=1, complete=False)
    assert names == [NAME + "/98", NAME + "/99"]

    names = hdf5.get_all_subgroup_names(FILENAME_FIX, EXAMPLE_OBJECT, depth=2, complete=False)
    assert names == [NAME + "/98/data", NAME + "/99/data"]

    names = hdf5.get_all_subgroup_names(FILENAME_FIX, EXAMPLE_OBJECT, depth=1, complete=True)
    assert names == [HALF_GROUP_NAME + "/98", HALF_GROUP_NAME + "/99"]

    assert not hdf5.get_all_subgroup_names(FILENAME_FIX, "NotExistingObject")

def test_load_all():
    """ test loading of several objects """
    with pytest.warns(UserWarning, match="Experimental parsing of attributes from group name"):
        objs = hdf5.load_all(ExampleObject, FILENAME_FIX)
        _check_loaded_objs(objs)

        objs = hdf5.load_all(ExampleObject, FILENAME_FIX, prefix_group_name=HALF_GROUP_NAME, used_default=False)
        _check_loaded_objs(objs)

        objs = hdf5.load_all(ExampleObject, FILENAME_FIX, index_range=[1])
        assert len(objs) == 1
        assert objs[0].name == NAME
        assert objs[0].value == VALUE

def _check_loaded_objs(objs):
    assert len(objs) == 2
    assert objs[0].name == NAME
    assert objs[0].value == 98
    assert objs[1].name == NAME
    assert objs[1].value == VALUE



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ObjectiveTerms """

from abc import abstractmethod
from warnings import warn

from .polynomial import _try_sorted, MIN_ABS_DIST
from .poly_binary import PolyBinary, get_all_reduction_penalty_terms
from .poly_ising import PolyIsing
from .objective import Objective
from .utils import variables


CONSTRAINT_TERMS_NAMES = "constraint_terms_names"
OBJECTIVE_TERMS = "objective_terms"
REDUCTION = "reduction"

ERROR_IMPL       = "Provide '{0}' on initialization or implement '_get_{0}' in inheriting 'ObjectiveTerms' class"
ERROR_NAMES      = "All constraint terms names should also be a key of the objective terms"
ERROR_MISMATCH   = "There is a mismatch between given and stored objective_terms names"
ERROR_TYPE       = "Cannot add PolyIsing and PolyBinary, make sure your objective terms are all of the same " \
                   "type by using the methods 'all_to_xxx'"
ERROR_DEGREE     = "Can only reduce to at least quadratic"
ERROR_REDUCTION  = "Reduction is currently only implemented for binary polynomials"
ERROR_OBJECTIVES = "Currently only one objective function is supported"
ERROR_INPUT      = "Either provide ObjectiveTerms objects or the constraint terms names separately"

WARNING_NOTHING     = "There is nothing in this ObjectiveTerms"
WARNING_DEPRECATION = "This way of instantiation is deprecated and will be removed soon, " \
                      "use method `get_from_instance` instead"
WARNING_SCALE       = "If the terms weights are given, the penalty scale is ignored"


class ObjectiveTerms(dict):
    """
    An objective term is a quadratic polynomial that belongs to a specific part of an objective function of a QUBO or
    an Ising problem. To distinguish the terms, they are named. Several objective terms are weighted and summed up to
    finally form the full objective polynomial.

    This class either serves as a container for pre-calculated objective terms
    or as the base class for the implementation of methods providing the objective terms.

    (It is an analogue to the ConstrainedObjective but with storing objective terms.)
    """

    def __init__(self, objective_terms=None, constraint_terms_names=None, name=None, instance=None):
        """
        initialize ObjectiveTerms object,
        either provide objective_terms to store them or
        provide instance for implemented subclass to construct the objective_poly and constraints from the instance

        :param (dict or None) objective_terms: mapping from names to objective terms
        :param (list[str] or None) constraint_terms_names: names of objective terms that correspond to constraints
        :param (str or None) name: identifying name to differ between several objective terms
        :param instance: instance object storing all data to construct the objective terms from
        """
        if instance:
            warn(WARNING_DEPRECATION, DeprecationWarning)
            self.constraint_terms_names = self._get_constraint_terms_names()
            self.name = name or self._get_name(instance)
            super().__init__(self._get_objective_terms(instance))
        else:
            self.constraint_terms_names = constraint_terms_names or []
            self.name = name
            super().__init__(objective_terms or {})

        self.check_consistency()

    def check_consistency(self):
        """ check if the data is consistent """
        if not set(self.constraint_terms_names) <= set(self.keys()):
            raise ValueError(ERROR_NAMES)
        if not self.constraint_terms_names and not self:
            warn(WARNING_NOTHING)

    def __eq__(self, other):
        if isinstance(other, ObjectiveTerms):
            if super().__eq__(other):
                return set(self.constraint_terms_names) == set(other.constraint_terms_names)
        return False

    def __ne__(self, other):
        # needs to be implemented otherwise would just use dict comparison
        return not self.__eq__(other)

    def __repr__(self):
        """ get nice string representation """
        return self.__class__.__name__ + f"({super().__repr__()}, {self.constraint_terms_names}, {repr(self.name)})"

    def __str__(self):
        """ get nice human-readable string representation of the objective terms """
        first_term = "min  P_{} * "
        other_terms = "\n     + P_{} * ".join(f"( {str(poly)} )" for poly in self.values())
        objective = (first_term + other_terms).format(*range(len(self)))
        domains = sorted(set(variables.to_domain_string(var, poly.__class__.__name__)
                             for poly in self.values() for var in poly.variables))
        domains = "\ns.t. " + ",\n     ".join(domains)
        return objective + domains

    def _check_names(self, names):
        if set(names) != set(self.keys()):
            raise ValueError(ERROR_MISMATCH)

    @classmethod
    def get_from_instance(cls, instance, name=None):
        """
        get the implemented inheriting ObjectiveTerms object based on the instance data

        :param (Instance or object) instance: An object of the implemented (inheriting) Instance class
        :param (str or None) name: the name of the resulting object
        :return: the initialized object
        """
        objective_terms = cls._get_objective_terms(instance)
        constraint_terms_names = cls._get_constraint_terms_names()
        name = name or cls._get_name(instance)
        return cls(objective_terms, constraint_terms_names, name)

    def reduce(self, max_degree=2, single_penalty_term=True, new_name=None, reduction_strategy=None):
        """
        reduce all appearing polynomials if the degrees are too large

        :param (int or None) max_degree: the maximal degree of the final polynomial,
                                         if None only the given reductions are applied
        :param (bool) single_penalty_term: True if all reductions shall be combined in a single penalty term
        :param (str or None) new_name: name of the new objective terms, by default the original name is used
        :param (str) reduction_strategy: the method how the reduction pairs are found, by default the fastest version
        :return: the reduced objective terms object
        """
        if max_degree < 2:
            raise ValueError(ERROR_DEGREE)
        if not all(isinstance(term, PolyBinary) for term in self.values()):
            raise ValueError(ERROR_REDUCTION)
        if all(term.degree <= max_degree for term in self.values()):
            return self

        new_terms = {}
        reductions = []
        for name, term in self.items():
            reduced_term, further_reductions = term.reduce(max_degree, use=reductions,
                                                           reduction_strategy=reduction_strategy)
            reductions += further_reductions
            new_terms[name] = reduced_term

        # apply reductions found in a later step also in already processed terms
        reductions = sorted(set(reductions))
        new_terms = {name: term.reduce(max_degree, use=reductions, reduction_strategy=reduction_strategy)[0]
                     for name, term in new_terms.items()}
        reduction_penalty_terms = get_all_reduction_penalty_terms(reductions)

        if not single_penalty_term:
            new_terms.update(reduction_penalty_terms)
            new_terms_names = self.constraint_terms_names + list(reduction_penalty_terms.keys())
        else:
            summed_reduction_penalty_term = sum(poly for poly in reduction_penalty_terms.values())
            new_terms[REDUCTION] = summed_reduction_penalty_term
            new_terms_names = self.constraint_terms_names + [REDUCTION]

        new_name = new_name or self.name
        return self.__class__(objective_terms=new_terms, constraint_terms_names=new_terms_names, name=new_name)

    def evaluate_terms(self, var_assignment):
        """
        get the values of all objective terms evaluated with the given variable assignment

        :param (dict) var_assignment: mapping from variables to values
        :return: the dictionary of objective terms names to values
        """
        return {name: objective_term.evaluate(var_assignment) for name, objective_term in self.items()}

    def check_validity(self, var_assignment):
        """
        check if the given variable assignment fulfills the condition
        that the value of all objective terms corresponding to constraints should be 0

        :param (dict) var_assignment: mapping from variables to values
        :return: True if all objective terms corresponding to constraints evaluate to 0
        """
        values = self.evaluate_terms(var_assignment)
        return all(values[name] == 0 for name in self.constraint_terms_names)

    def get_terms_weights_string(self, terms_weights):
        """
        get a string representation of the terms weights,
        the keys need to be the same as the objective term names
        the ordering in the string is defined by the ordering in the terms weights dictionary
        (e.g. 'myConstraint1.0_mySecondConstraint1.5')

        :param (dict) terms_weights: mapping of objective terms names to values
        :return: the name consisting of the objective terms names with the corresponding values
        """
        self._check_names(terms_weights)
        return get_string(terms_weights)

    def get_weighted_sum_polynomial(self, terms_weights):
        """
        get the weighted sum of the objective terms as a polynomial

        :param (dict) terms_weights: mapping of objective terms names to values
        :return: the weighted sum of the objective terms
        """
        self._check_names(terms_weights)
        try:
            return sum(value * self[name] for name, value in terms_weights.items())
        except ValueError as ve:
            raise ValueError(ERROR_TYPE) from ve

    def get_objective(self, terms_weights=None, penalty_scale=1, name=None):
        """
        get the objective with a polynomial objective function given by the weighted sum of the objective terms,

        if the resulting solutions are not feasible, i.e., they do not fulfill the constraints,
        increase the penalty scale or increase the single terms_weights correspondingly

        :param (dict) terms_weights: mapping of objective terms names to values
        :param (numbers.Real) penalty_scale: the scaling factor which is applied to the weights of terms corresponding
                                             to constraints, only if no terms_weights are given
        :param (str or None) name: name of the constructed objective, by default generated from terms weights
        :return: the Objective with the weighted sum of the objective terms as objective function
        """
        if terms_weights and penalty_scale != 1:
            warn(WARNING_SCALE)
        terms_weights = terms_weights or self.get_default_terms_weights(penalty_scale)
        poly = self.get_weighted_sum_polynomial(terms_weights)
        name = name or self.get_default_objective_name(terms_weights)
        return Objective(polynomial=poly, name=name)

    def get_default_terms_weights(self, penalty_scale=1, use_naive_bounds=False):
        """
        get a default weighting of the objective terms to create the objective,
        can be overwritten in subclass implementations to provide more meaningful weights

        :param (numbers.Real) penalty_scale: the scaling factor which is applied to the weights of terms corresponding
                                             to constraints
        :param (bool) use_naive_bounds: whether to use naive bound or not
        :return: the default terms weights
        """
        scaled_terms_weights = {name: penalty_scale if name in self.constraint_terms_names else 1 for name in self}
        if not use_naive_bounds:
            return scaled_terms_weights

        naive_terms_weights = get_naive_terms_weights(self)
        return {name: value * scaled_terms_weights[name] for name, value in naive_terms_weights.items()}

    def get_default_objective_name(self, terms_weights):
        """
        get the default name of the constructed objective

        :param (dict) terms_weights: mapping of objective terms names to values
        :return: the default name of the objective
        """
        prefix = "" if not self.name else f"{self.name}/"
        return prefix + self.get_terms_weights_string(terms_weights)

    def get_all_variables(self):
        """
        get all variables from all polynomials

        :return: the sorted list of all appearing variables
        """
        all_vars = set().union(*(term.variables for term in self.values()))
        return _try_sorted(all_vars)

    def combine_prefixes(self, *prefixes, new_name=None):
        """
        sum up those objective terms whose names share the given prefixes

        :param (str) prefixes: the prefixes of the objective terms names that shall be combined
        :param (str or None) new_name: name of the new objective terms, by default the original name is used
        :return: the new ObjectiveTerms object with fewer terms
        """
        new_terms = dict(self)
        new_names = set(self.constraint_terms_names)
        for prefix in prefixes:
            combine = {name: term for name, term in self.items() if name.startswith(prefix)}
            if combine:
                for name in combine.keys():
                    del new_terms[name]
                    if name in new_names:
                        new_names.remove(name)
                        new_names.add(prefix)
                new_terms[prefix] = sum(combine.values())

        new_name = new_name or self.name
        return self.__class__(objective_terms=new_terms, constraint_terms_names=sorted(new_names), name=new_name)

    def all_to_binary(self, new_name=None):
        """
        convert all objective terms to PolyBinary

        :param (str or None) new_name: name of the new objective terms, by default the original name is used
        :return: the new ObjectiveTerms with all terms as binary polynomials
        """
        new_name = new_name or self.name
        new_terms = {name: PolyBinary.from_unknown_poly(poly) for name, poly in self.items()}
        return self.__class__(objective_terms=new_terms,
                              constraint_terms_names=self.constraint_terms_names,
                              name=new_name)

    def all_to_ising(self, inverted=False, new_name=None):
        """
        convert all objective terms to PolyIsing

        :param (bool) inverted: if True, convert to inverted Ising
        :param (str or None) new_name: name of the new objective terms, by default the original name is used
        :return: the new ObjectiveTerms with all terms as Ising polynomials
        """
        new_name = new_name or self.name
        new_terms = {name: PolyIsing.from_unknown_poly(poly, inverted=inverted) for name, poly in self.items()}
        return self.__class__(objective_terms=new_terms,
                              constraint_terms_names=self.constraint_terms_names,
                              name=new_name)

    @staticmethod
    @abstractmethod
    def _get_constraint_terms_names():
        """
        get the names of the objective terms whose value must be zero to have the corresponding constraints fulfilled,
        which need to be a subset of all objective terms names,
        needs to be implemented in subclasses for automatic generation

        :return: the list of objective terms names representing the constraints
        """
        raise NotImplementedError(ERROR_IMPL.format(CONSTRAINT_TERMS_NAMES))

    @staticmethod
    @abstractmethod
    def _get_objective_terms(instance):
        """
        get the objective terms from the instance data,
        needs to be implemented in subclasses for automatic generation

        :param instance: instance object storing all data to construct the objective terms from
        :return: the dictionary with names to objective terms (Polynomials)
        """
        raise NotImplementedError(ERROR_IMPL.format(OBJECTIVE_TERMS))

    @staticmethod
    def _get_name(instance):  # pylint: disable=unused-argument
        """
        get the name from the instance data,
        can be implemented in subclasses for automatic generation, is not enforced

        :param instance: instance object storing all data to construct the name from
        :return: the name of the object
        """
        return None


def get_string(terms_weights):
    """
    get a string concatenating the given dictionary of terms weights,
    the ordering in the string is defined by the ordering in the terms weights dictionary
    (e.g. myConstraint1.0_mySecondConstraint1.5)

    :param (dict) terms_weights: mapping of objective terms names to values
    :return: the string consisting of the objective terms names with the corresponding values
    """
    return '_'.join(name + f'{weight:e}' for name, weight in terms_weights.items())

def get_naive_terms_weights(objective_terms, constraint_terms_names=None, gap=1):
    """
    get naive weights for the objective terms, derived from naive bounds on the objective and the constraint terms

    :param (dict) objective_terms: the objective_terms to calculate the weights for
    :param (list) constraint_terms_names: the names of the constraint terms to weight the objective against
    :param (float) gap: the difference between the objective and constraint terms
    :return: the naive terms weights
    """
    if isinstance(objective_terms, ObjectiveTerms):
        constraint_terms_names = objective_terms.constraint_terms_names
    if not constraint_terms_names:
        raise ValueError(ERROR_INPUT)

    # get the objective function
    objective_functions = set(objective_terms).difference(constraint_terms_names)
    if not objective_functions:
        # if there is no objective function, all constraint terms can be weighted equally
        return {name: 1 for name in constraint_terms_names}
    if len(objective_functions) > 1:
        raise ValueError(ERROR_OBJECTIVES)
    objective_name = objective_functions.pop()

    # get naive bound on the benefit in the objective
    obj_lower_bound = objective_terms[objective_name].naive_lower_bound
    obj_upper_bound = objective_terms[objective_name].naive_upper_bound
    objective_bound = obj_upper_bound - obj_lower_bound + gap

    # for constraint terms with non-integer coefficients, a normalizing factor needs to be applied
    normalizing_factors = {name: poly.coefficients_info[MIN_ABS_DIST] if not poly.has_int_coefficients() else 1
                           for name, poly in objective_terms.items()}

    return {name: objective_bound / normalizing_factors[name] if name in constraint_terms_names else 1
            for name in objective_terms.keys()}



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the HardwareAdjacency """

from quark import HardwareAdjacency
from . import add, hdf5_datasets


EDGES = "edges"


add.save_load_exists(HardwareAdjacency)

@add.static_method(HardwareAdjacency)
def get_identifying_attributes():
    """ the attributes that identify an object of the class """
    return ["name"]

@add.self_method(HardwareAdjacency)
def write_hdf5(hardware_adjacency, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (HardwareAdjacency) hardware_adjacency: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    hdf5_datasets.write_dataset(group, EDGES, dataset=hardware_adjacency)


@add.static_method(HardwareAdjacency)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    return hdf5_datasets.read_datasets(group, EDGES)



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the Embedding """

from quark import Embedding
from . import add, hdf5_datasets


VAR_NODES_MAP      = "var_nodes_map"
VAR_EDGES_MAP      = "var_edges_map"
COUPLING_EDGES_MAP = "coupling_edges_map"
EDGES              = "edges"


add.save_load_exists(Embedding)

@add.static_method(Embedding)
def get_identifying_attributes():
    """ the attributes that identify an object of the class """
    return ["name", "index"]

@add.self_method(Embedding)
def write_hdf5(embedding, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (Embedding) embedding: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    edge_map = {edge: i for i, edge in enumerate(embedding.edges)}
    # replacing edges by their index in the edge list to be able to store it
    coupling_edges_map = {var: [edge_map[edge] for edge in edges]
                          for var, edges in embedding.coupling_edges_map.items()}
    var_edges_map = {var: [edge_map[edge] for edge in edges] for var, edges in embedding.var_edges_map.items()}

    hdf5_datasets.write_datasets(group, embedding, VAR_NODES_MAP, EDGES)
    hdf5_datasets.write_datasets(group, var_edges_map=var_edges_map, coupling_edges_map=coupling_edges_map)

@add.static_method(Embedding)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    var_nodes_map = hdf5_datasets.read_dataset(group, VAR_NODES_MAP)
    init_kwargs = {VAR_NODES_MAP: {var: list(nodes) for var, nodes in var_nodes_map.items()}}

    edges = hdf5_datasets.read_dataset(group, EDGES, check_existence=False)
    if edges is not None:
        var_edges_map = _get_map(VAR_EDGES_MAP, group, edges)
        coupling_edges_map = _get_map(COUPLING_EDGES_MAP, group, edges)
        init_kwargs.update(var_edges_map=var_edges_map, coupling_edges_map=coupling_edges_map)
    return init_kwargs


def _get_map(name, group, edges):
    """ get and convert the corresponding dataset """
    edge_map = hdf5_datasets.read_dataset(group, name, check_existence=False)
    if edge_map is not None:
        edge_map = {var: [tuple(edges[i]) for i in edge_indices] for var, edge_indices in edge_map.items()}
    return edge_map



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for PolyIsing """

from collections import Counter
from functools import wraps
from math import copysign

from .polynomial import Polynomial
from . import poly_binary


ADD      = "add"
MULTIPLY = "multiply"

ERROR_INVERTED = "Cannot {} two PolyIsings with different inversions"
ERROR_BINARY   = "Cannot {} PolyIsing and PolyBinary"
ERROR_INVERT   = "Choose either to set certain inversion of or to invert the possible PolyIsing"


def _call_super_and_set_inverted(decorated_func):
    """
    a decorator which takes the returned Polynomial from the decorated method
    and returns a PolyIsing instead with the correctly set attribute '_inverted'

    :return: the new function
    """
    @wraps(decorated_func)
    def wrapper(self, *args, **kwargs):
        """
        the wrapper function executes the decorated function, executes its super function and, if the resulting
        polynomial is a PolyIsing, sets the inverted flag
        """
        # call decorated function
        decorated_func(self, *args, **kwargs)
        # get the super function
        super_func = getattr(super(PolyIsing, self), decorated_func.__name__)
        # get the resulting polynomial
        result_poly = super_func(*args, **kwargs)
        if isinstance(result_poly, PolyIsing):
            # set the correct inverted flag if it is an Ising polynomial
            result_poly._inverted = self.is_inverted()  #pylint: disable=protected-access
        return result_poly

    return wrapper


class PolyIsing(Polynomial):
    """
    An Ising polynomial is a special polynomial, where the variables can only take -1 or 1.
    Since (-1)^2 = 1^2 = 1, we can simplify the Polynomial by removing all even exponents,
    e.g.

        * (0, 0, 1, 1, 1): 1 -> (1,): 1,
        * (2, 2, 2, 2): 2    -> (): 2.

    We have the following conventions:

        *  PolyIsing == PolyIsing <=> poly == poly and inverted == inverted,
        *  PolyIsing == Polynomial <=> poly == poly,
        *  PolyIsing +/* Polynomial = PolyIsing,
        *  Polynomial +/* PolyIsing = Polynomial + Warning,
        *  PolyIsing(inverted=True) +/* PolyIsing(inverted=False) = Error,
        *  PolyIsing +/* PolyBinary = Error,
        *  copy/remove_zero_coefficients/replace_variables/get_rounded(PolyIsing) = PolyIsing
                (with correctly set inverted attribute),
        *  evaluate(PolyIsing, {var : Number}) = PolyIsing/Number,
        *  evaluate(PolyIsing, {var : Polynomial}) = Polynomial,
        *  evaluate(PolyIsing, {var : PolyIsing}) = Error,
        *  PolyBinary == PolyIsing = False.
    """

    def __init__(self, d=None, inverted=False, variable_tuples_already_formatted=False):
        """
        initialize PolyIsing object

        :param (dict) d: mapping from tuples of variables to coefficients
        :param (bool) variable_tuples_already_formatted: if True, d will be used as passed to save time on sorting,
                                                         use if d is, e.g., a dictionary of an existing Polynomial,
                                                         otherwise every key in d will be sorted
        :param (bool) inverted: flag that defines the relation to the binary format:
                                False : 0 <-> -1, 1 <-> 1
                                True  : 0 <-> 1, 1 <-> -1
        """
        self._inverted = inverted
        super().__init__(d, variable_tuples_already_formatted)

    def _format_variable_tuple(self, var_tuple):
        """
        transform the tuple of variables into the standard format, also checks the type of the variables,
        extends 'Polynomial._format_variable_tuple',
        since we have s * s = 1  for all s in {-1, +1}, all even powers can be removed
        """
        var_tuple = super()._format_variable_tuple(var_tuple)
        var_powers = Counter(var_tuple)
        if not len(var_powers) == len(var_tuple):
            var_tuple = tuple(var for var, power in var_powers.items() if power % 2 > 0)
        return var_tuple

    def __eq__(self, other):
        """
        check equality of polynomials, where PolyIsing is never equal to a PolyBinary,
        extends 'Polynomial.__eq__'
        """
        if isinstance(other, poly_binary.PolyBinary):
            return False
        if super().__eq__(other):
            if isinstance(other, PolyIsing):
                # if compared to another PolyIsing also check the inverted attribute
                return self.is_inverted() == other.is_inverted()
            # if PolyIsing is compared to Polynomial we ignore the attribute 'inverted'
            return True
        return False

    def _check_different(self, poly2, func_name):
        """ check if the type of the second Polynomial is feasible for an operation """
        if isinstance(poly2, PolyIsing) and poly2.is_inverted() != self.is_inverted():
            raise ValueError(ERROR_INVERTED.format(func_name))
        if isinstance(poly2, poly_binary.PolyBinary):
            raise ValueError(ERROR_BINARY.format(func_name))

    @_call_super_and_set_inverted
    def __add__(self, poly2):
        """
        add another polynomial or a scalar,
        extends 'Polynomial.__add__' since PolyBinary cannot be added to PolyIsing
        """
        self._check_different(poly2, ADD)

    @_call_super_and_set_inverted
    def __mul__(self, poly2):
        """
        multiply with another polynomial or a scalar,
        extends 'Polynomial.__mul__' since PolyBinary cannot be multiplied with PolyIsing
        """
        self._check_different(poly2, MULTIPLY)

    # pylint: disable=multiple-statements  # Better readability
    @_call_super_and_set_inverted
    def copy(self): pass

    @_call_super_and_set_inverted
    def remove_zero_coefficients(self): pass

    @classmethod
    def _is_valid_var_assignment(cls, var_assignment):
        """
        check whether the given input can be assigned to the variable,
        that means it is a Polynomial, PolyBinary, SCIP Expr or valid coefficient,
        extends 'Polynomial._is_valid_var_assignment' to not allow for PolyBinary
        """
        return not isinstance(var_assignment, PolyIsing) and super()._is_valid_var_assignment(var_assignment)

    @_call_super_and_set_inverted
    def evaluate(self, var_assignments, keep_poly=False): pass

    @_call_super_and_set_inverted
    def replace_variables(self, replacement): pass

    @_call_super_and_set_inverted
    def round(self, decimal_cap=None): pass

    # additional methods for Ising polynomials

    @classmethod
    def from_unknown_poly(cls, poly, inverted=None, invert=False):
        """
        convert a polynomial of unknown type into a PolyIsing

        :param (Polynomial) poly: the polynomial of unknown type
        :param (bool or None) inverted: if set to False the resulting PolyIsing will be not be an inverted one
                                        if set to True the resulting PolyIsing will be an inverted one
        :param (bool) invert: if set to True the resulting PolyIsing will be inverted
                              meaning if it was non-inverted PolyIsing before it will be inverted afterwards
                              resp. if it was an inverted PolyIsing before it will not be inverted anymore
                              (cannot be set if inverted is set)
        :return: the corresponding Ising polynomial
        """
        if not (inverted is None or not invert):
            raise ValueError(ERROR_INVERT)

        if isinstance(poly, poly_binary.PolyBinary):
            return poly.to_ising(inverted or invert)
        if isinstance(poly, cls):
            if inverted is not None and poly.is_inverted() != inverted:
                return poly.invert()
            if inverted is None and invert:
                return poly.invert()
            return poly
        return cls(poly, inverted=inverted or invert)

    def preprocess_minimize(self, unambitious=False):
        """
        simplify the polynomial by setting some variables to the values which are guaranteed to appear in an optimal
        solution when considering minimization

        :param (bool) unambitious:
            if True, only those variables will be set where their assignment is unambitious, that means, in all optimal
            solutions, they have the same value,
            if False, preprocess the variable to the straightforward value whether it might also get another value in an
            optimal solution, the objective values remains the same
        :return: the new reduced Polynomial and the assignment of the eliminated variables with variable to value
        """
        return super().preprocess(preprocess_rule, unambitious)

    def is_inverted(self):
        """ check if the PolyIsing is inverted """
        return self._inverted

    def invert(self):
        """ interchange 1 and -1 in solutions of the PolyIsing by applying the mapping s -> -s """
        result = self.affine_transform(-1)
        return PolyIsing(result, not self._inverted)

    def to_binary(self):
        """
        convert the Ising polynomial to the binary format
        by convention it is always from the not inverted ising,
        therefore if inverted is
            - False, we use the conversion -1 -> 0, 1 -> 1 with s = 2x - 1,
            - True, we use the conversion -1 -> 1, 1 -> 0 with s = -2x + 1

        :return: the corresponding PolyBinary
        """
        converted = self.affine_transform(-2, 1) if self.is_inverted() else self.affine_transform(2, -1)
        return poly_binary.PolyBinary(converted)


def preprocess_rule(var_sigmas, coeff, unambitious=False):
    """
    very simple preprocessing function:
    if the absolute weight of the variable extends all influences from the outside
    (i.e. the sum of all absolute strengths on couplings including this variable = positive sigma - negative sigma)
    the variable can be assigned in advance, such that the product of var_assignment * weight is negative
    therefore negative sign of the coefficient

    :param (tuple) var_sigmas: sigma values (combined incoming strengths) of a variable of the polynomial
    :param (Real) coeff: coefficient on the linear term of the variable
    :param (bool) unambitious: if True only those variables will be set where their assignment is unambitious,
                               that means, in all optimal solutions, they have the same value,
                               if False, preprocess the variable to the straightforward value whether it might also
                               get another value in an optimal solution, the objective values remains the same
    :return: the solution value or None if the variable cannot be preprocessed
    """
    if unambitious and abs(coeff) == var_sigmas[0] - var_sigmas[1]:
        return None
    if abs(coeff) >= var_sigmas[0] - var_sigmas[1]:
        return -copysign(1, coeff)
    return None



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the Solution """

from numbers import Integral

from quark import Solution
from . import add, hdf5_attributes, hdf5_datasets
from ._variable_mapping import write_variables, read_variables


ATTRS_IMPORTANT   = ["objective_value", "solving_success", "solving_status", "solving_time"]
ATTRS_UNIMPORTANT = ["total_time", "dual_gap", "dual_bound"]
ATTRS = ATTRS_IMPORTANT + ATTRS_UNIMPORTANT
VAR_ASSIGNMENTS = "var_assignments"


add.save_load_exists(Solution)

@add.static_method(Solution)
def get_identifying_attributes():
    """ the attributes that identify an object of the class """
    return ["name"]

@add.self_method(Solution)
def write_hdf5(solution, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (Solution) solution: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    hdf5_attributes.write_attributes(group, solution, *ATTRS)
    if all(isinstance(var, Integral) for var in solution.keys()):
        var_assignments = solution
    else:
        # the variables are more complicated and need to be stored differently
        var_assignments = solution.replace_variables({var: index for index, var in enumerate(solution.keys())})
        write_variables(solution.keys(), group)
    hdf5_datasets.write_dataset(group, VAR_ASSIGNMENTS, dataset=var_assignments)


@add.static_method(Solution)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    init_kwargs = hdf5_attributes.read_attributes(group, *ATTRS_IMPORTANT)
    init_kwargs.update(hdf5_attributes.read_attributes(group, *ATTRS_UNIMPORTANT, check_existence=False))

    var_assignments = hdf5_datasets.read_dataset(group, VAR_ASSIGNMENTS)
    composite_variables = read_variables(group)
    if composite_variables:
        var_assignments = {composite_variables[index]: value for index, value in var_assignments.items()}

    init_kwargs[VAR_ASSIGNMENTS] = var_assignments
    return init_kwargs



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the configuration base class """

import os
import pytest

from quark.utils.config import Config, dict_to_string, _to_string
from quark.testing import ExampleConfig


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILENAME_TMP = TEST_DIR + "test_config.cfg"

DEFAULT_CONFIG = {"float_option": 20.0,
                  "int_option": 10,
                  "list_option": None,
                  "str_option": "some_string",
                  "dict_option": {"default": 0}}


def test_name():
    """ Test if the name gets returned correctly """
    conf = ExampleConfig()
    assert conf.get_name() == "ExampleConfig"

def test_config():
    """ Test the example configuration file, check if parameters get set correctly """
    config = ExampleConfig(1.0, int_option=10)
    assert config.int_option == 10
    assert config["float_option"] == 1.0
    config.int_option = 5
    assert config["int_option"] == 5
    del config["int_option"]
    assert config.int_option is None
    assert config.str_option is None
    config.update(str_option="test", dict_option={"test": 123})
    assert config.str_option == "test"
    assert config["dict_option"] == {"test": 123}
    assert config.int_option is None
    config.str_option = None
    assert config["str_option"] is None
    config.set_defaults(override=False)
    assert config.float_option == 1.0
    assert config.int_option == 10
    assert config["str_option"] == "some_string"
    assert config.dict_option == {"test": 123}
    config.set_defaults()
    assert config.float_option == 20.0
    assert config.int_option == 10
    assert config["str_option"] == "some_string"
    assert config.dict_option == {"default": 0}
    config.str_option = "1"
    assert config.str_option != 1
    assert config.str_option == "1"

    errors = [lambda x: config[x],
              lambda x: config.x,
              lambda x: config.update(**{x: 1}),
              config.__getattr__,
              config.__getitem__,
              config.__delattr__,
              config.__delitem__]
    for call in errors:
        with pytest.raises(AttributeError):
            call("non_existing_key")

    config.num_reads = "string instead of number"
    assert config.num_reads == "string instead of number"

    config.int_option = dict({"test.": 123})
    assert isinstance(config.int_option, dict)

def test_get_options_dict():
    """ Test if the options dict returns contains correct parameters """
    config = ExampleConfig(set_defaults=True)
    options_dict = config.get_options_dict()
    for key, option in DEFAULT_CONFIG.items():
        assert options_dict[key]["default"] == option

    with pytest.raises(NotImplementedError):
        Config.get_options_dict()

def test_io():
    """ test read and write of the example Config """
    config_1 = ExampleConfig(float_option=100.50, int_option=5, str_option="hello_world")
    config_2 = ExampleConfig(float_option=1.0, int_option=1300, set_defaults=True)
    assert config_1.float_option != config_2.float_option
    config_1.save(FILENAME_TMP)
    config_1_loaded = ExampleConfig.load(filename=FILENAME_TMP)
    assert config_1 == config_1_loaded

    config_2.save(FILENAME_TMP)
    config_2_loaded = ExampleConfig.load(FILENAME_TMP)
    assert config_1_loaded != config_2_loaded
    assert config_2 == config_2_loaded

    config_broken = ExampleConfig(float_option="10aa0")
    config_broken.save(FILENAME_TMP)
    with pytest.raises(TypeError, match="The value is of type '<class 'str'>'"):
        config_broken.load(FILENAME_TMP)

    config_broken = ExampleConfig(str_option=10, dict_option=[])
    config_broken.save(FILENAME_TMP)
    with pytest.raises(TypeError, match="The value is of type '<class 'list'>'"):
        config_broken.load(FILENAME_TMP)

    os.remove(FILENAME_TMP)

def test_to_string():
    """ test conversion to strings """
    config = ExampleConfig(float_option=1.23,
                           int_option=1024,
                           list_option=[1.1, 2.2],
                           set_defaults=True)
    expected_string = "_float_option1.230000e+00"
    expected_string += "_int_option1.024000e+03"
    expected_string += "_list_option_1.100000e+00_2.200000e+00"
    expected_string += "_str_option_some_string"
    expected_string += "_dict_option_default0.000000e+00"
    assert config.to_string(max_depth=1) == expected_string

    with pytest.raises(ValueError, match="Level of nested dictionaries exceeds specified maximal depth 0"):
        config.to_string(max_depth=0)

def test_update():
    """ test of the update method """
    config = ExampleConfig(set_defaults=True)
    assert config["float_option"] == 20.0
    config.update(float_option=50.33)
    assert config["float_option"] == 50.33
    with pytest.raises(AttributeError):
        config.update(not_existing_argument=100)

def test_class_methods():
    """ test of the class methods """
    conf = ExampleConfig(set_defaults=True)
    filename_not_exist = "nowhere"
    file_not_exist = TEST_DIR + filename_not_exist
    with pytest.raises(ValueError, match=f"Configuration file '.*{filename_not_exist}' does not exists"):
        conf.load(filename=file_not_exist)

    with open(FILENAME_TMP, mode="a", encoding="utf-8") as txt_file:
        with pytest.raises(ValueError, match="Missing section 'ExampleConfig' in configuration file"):
            conf.load(FILENAME_TMP)

        txt_file.write("[ExampleConfig]\n")

    with pytest.raises(ValueError, match="Missing option 'float_option' in configuration file"):
        conf.load(FILENAME_TMP)

    assert conf.get_defaults() is not None
    assert conf.get_help() is not None

    os.remove(FILENAME_TMP)


STR = "text"
INT = 1024
FLOAT = 1.23
LIST = [1.1, 2.2]
BOOL = True
SUBSUB_DICT = dict({"a": 1, "b": 1.1})
SUB_DICT1 = dict({"answer": "yes", "subsubdict": SUBSUB_DICT})
SUB_DICT2 = dict({"A": 2, "B": 2.2, "to_omit_in_subdict": "some/complicated/path/to/omit/"})
DICTIONARY = {"str"      : STR,
              "int"      : INT,
              "bool"     : BOOL,
              "float"    : FLOAT,
              "subdict1" : SUB_DICT1,
              "subdict2" : SUB_DICT2,
              "list"     : LIST,
              "to_omit"  : "some/complicated/path/to/omit/"}

def test_value_to_string():
    """ test value to string conversion """
    assert _to_string(STR) == STR
    assert _to_string(FLOAT) == "1.230000e+00"
    assert _to_string(INT) == "1.024000e+03"
    with pytest.raises(ValueError, match="Values must be of one of the types int, float, str or bool, not 'NoneType'"):
        _to_string(None)

def test_dict_to_string():
    """ test conversion feasibility check """
    exp_string = "_str_text_int1.024000e+03_boolTrue_float1.230000e+00" \
                 "_subdict1_answer_yes_subsubdict_a1.000000e+00_b1.100000e+00_subdict2_A2.000000e+00_B2.200000e+00" \
                 "_list_1.100000e+00_2.200000e+00"
    string = dict_to_string(DICTIONARY, max_depth=5, omit_keys=["to_omit", "to_omit_in_subdict"])
    assert string == exp_string

    FAULTY_DICT = {1: "text"}
    with pytest.raises(ValueError, match="Keys must be of type str not 'int'"):
        dict_to_string(FAULTY_DICT, max_depth=4)

    FAULTY_DICT = {"text": None}
    with pytest.raises(ValueError, match="of one of the types int, float, str, bool, list or dict, not 'NoneType'"):
        dict_to_string(FAULTY_DICT, max_depth=4)
    with pytest.raises(ValueError, match="Level of nested dictionaries exceeds specified maximal depth 1"):
        dict_to_string(DICTIONARY, max_depth=1)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the ConstrainedObjective by inherited example and without """

import itertools
import pytest

from quark import VariableMapping, PolyBinary, ConstraintBinary, ConstrainedObjective
from quark.constraint_binary import get_reduction_constraints
from quark.testing import ExampleConstrainedObjective
from quark.testing.example_objective_terms import X, ONE_COLOR_PER_NODE, COLORED_EDGES


OBJECTIVE_POLY_FLAT_STR = "+1 x0 x3 +1 x0 x6 +1 x0 x12 +1 x1 x4 +1 x1 x7 +1 x1 x13 +1 x2 x5 " \
                          "+1 x2 x8 +1 x2 x14 +1 x3 x6 +1 x4 x7 +1 x5 x8 +1 x6 x9 +1 x6 x12 " \
                          "+1 x7 x10 +1 x7 x13 +1 x8 x11 +1 x8 x14 +1 x9 x12 +1 x10 x13 +1 x11 x14"

Y, R = "y", "reduction"


def test_init(constrained_objective, exp_constrained_objective):
    """ test constrained objective initialization """
    assert constrained_objective == exp_constrained_objective

def test_repr(constrained_objective):
    """ test string representation """
    constrained_objective.name = "test"
    exp_str = "ExampleConstrainedObjective(" + repr(constrained_objective.objective_poly) + ", " \
              "{'one_color_per_node_1': " + repr(constrained_objective['one_color_per_node_1']) + ", " \
               "'one_color_per_node_2': " + repr(constrained_objective['one_color_per_node_2']) + ", " \
               "'one_color_per_node_3': " + repr(constrained_objective['one_color_per_node_3']) + ", " \
               "'one_color_per_node_4': " + repr(constrained_objective['one_color_per_node_4']) + ", " \
               "'one_color_per_node_5': " + repr(constrained_objective['one_color_per_node_5']) + "}, " \
              "'test')"
    assert repr(constrained_objective) == exp_str

def test_str(constrained_objective):
    """ test string representation """
    exp_str = "min  +1 x_1_blue x_2_blue +1 x_1_blue x_3_blue +1 x_1_blue x_5_blue " \
                   "+1 x_1_green x_2_green +1 x_1_green x_3_green +1 x_1_green x_5_green " \
                   "+1 x_1_red x_2_red +1 x_1_red x_3_red +1 x_1_red x_5_red " \
                   "+1 x_2_blue x_3_blue +1 x_2_green x_3_green +1 x_2_red x_3_red " \
                   "+1 x_3_blue x_4_blue +1 x_3_blue x_5_blue +1 x_3_green x_4_green " \
                   "+1 x_3_green x_5_green +1 x_3_red x_4_red +1 x_3_red x_5_red " \
                   "+1 x_4_blue x_5_blue +1 x_4_green x_5_green +1 x_4_red x_5_red\n" \
              "s.t. 1 == +1 x_1_blue +1 x_1_green +1 x_1_red,\n" \
              "     1 == +1 x_2_blue +1 x_2_green +1 x_2_red,\n" \
              "     1 == +1 x_3_blue +1 x_3_green +1 x_3_red,\n" \
              "     1 == +1 x_4_blue +1 x_4_green +1 x_4_red,\n" \
              "     1 == +1 x_5_blue +1 x_5_green +1 x_5_red,\n" \
              "     x_*_* in {0, 1}"
    assert str(constrained_objective) == exp_str

    # make an inequality
    constrained_objective[ONE_COLOR_PER_NODE + "_2"].lower_bound = -1
    exp_str = "min  +1 x_1_blue x_2_blue +1 x_1_blue x_3_blue +1 x_1_blue x_5_blue " \
                   "+1 x_1_green x_2_green +1 x_1_green x_3_green +1 x_1_green x_5_green " \
                   "+1 x_1_red x_2_red +1 x_1_red x_3_red +1 x_1_red x_5_red " \
                   "+1 x_2_blue x_3_blue +1 x_2_green x_3_green +1 x_2_red x_3_red " \
                   "+1 x_3_blue x_4_blue +1 x_3_blue x_5_blue +1 x_3_green x_4_green " \
                   "+1 x_3_green x_5_green +1 x_3_red x_4_red +1 x_3_red x_5_red " \
                   "+1 x_4_blue x_5_blue +1 x_4_green x_5_green +1 x_4_red x_5_red\n" \
              "s.t.  1 == +1 x_1_blue +1 x_1_green +1 x_1_red,\n" \
              "      1 == +1 x_3_blue +1 x_3_green +1 x_3_red,\n" \
              "      1 == +1 x_4_blue +1 x_4_green +1 x_4_red,\n" \
              "      1 == +1 x_5_blue +1 x_5_green +1 x_5_red,\n" \
              "     -1 <= +1 x_2_blue +1 x_2_green +1 x_2_red <= 1,\n" \
              "     x_*_* in {0, 1}"
    assert str(constrained_objective) == exp_str

    # flat polynomial, only one equality
    flat_poly = PolyBinary.read_from_string(OBJECTIVE_POLY_FLAT_STR)
    flat_constraint = ConstraintBinary(PolyBinary({(1,2) : 1}), 0, 0)
    constrained_objective = ConstrainedObjective(flat_poly, {"flat": flat_constraint})
    exp_str = "min  +1 x0 x3 +1 x0 x6 +1 x0 x12 +1 x1 x4 +1 x1 x7 +1 x1 x13 +1 x2 x5 +1 x2 x8 +1 x2 x14 +1 x3 x6 " \
                   "+1 x4 x7 +1 x5 x8 +1 x6 x9 +1 x6 x12 +1 x7 x10 +1 x7 x13 +1 x8 x11 +1 x8 x14 +1 x9 x12 " \
                   "+1 x10 x13 +1 x11 x14\n" \
              "s.t. 0 == +1 x1 x2,\n" \
              "     x* in {0, 1}"
    assert str(constrained_objective) == exp_str

    # polynomial with string variables, only inequalities
    str_poly = PolyBinary({('a', 'b'): 1, ('b', 'c'): 1})
    flat_constraint1 = ConstraintBinary(PolyBinary({('a',): 1, ('c',): 1}), 0, 1)
    flat_constraint2 = ConstraintBinary(PolyBinary({('a',): -1, ('c',): 1}), -1, 0)
    constrained_objective = ConstrainedObjective(str_poly, {"flat1": flat_constraint1, "flat2": flat_constraint2})
    exp_str = "min  +1 a b +1 b c\n" \
              "s.t.  0 <= +1 a +1 c <= 1,\n" \
              "     -1 <= -1 a +1 c <= 0,\n" \
              "     a in {0, 1},\n" \
              "     b in {0, 1},\n" \
              "     c in {0, 1}"
    assert str(constrained_objective) == exp_str

    # polynomial, no constraints
    poly = PolyBinary({((X, 1, 2), (Y, 0)): 1, ((X, 1, 2), (Y, 1)): 1})
    constrained_objective = ConstrainedObjective(poly)
    exp_str = "min  +1 x_1_2 y_0 +1 x_1_2 y_1\n" \
              "s.t. x_*_* in {0, 1},\n" \
              "     y_* in {0, 1}"
    assert str(constrained_objective) == exp_str

def test_get_from_instance(instance, constrained_objective):
    """ test different initialization methods """
    constrained_objective_get = ExampleConstrainedObjective.get_from_instance(instance)
    assert constrained_objective_get == constrained_objective

    with pytest.warns(DeprecationWarning, match="This way of instantiation is deprecated"):
        constrained_objective_depr = ExampleConstrainedObjective(instance=instance)
    assert constrained_objective_depr == constrained_objective

def test_get_all_variables(constrained_objective):
    """ test getting all variables """
    exp_variables = [(X, 1, "blue"), (X, 1, "green"), (X, 1, "red"), (X, 2, "blue"), (X, 2, "green"), (X, 2, "red"),
                     (X, 3, "blue"), (X, 3, "green"), (X, 3, "red"), (X, 4, "blue"), (X, 4, "green"), (X, 4, "red"),
                     (X, 5, "blue"), (X, 5, "green"), (X, 5, "red")]
    assert constrained_objective.get_all_variables() == exp_variables

    # introduce constraint with wrong variables manually
    constrained_objective["wrong"] = ConstraintBinary(PolyBinary({("x",): 1, ("y",): 1}), 0, 1)
    with pytest.raises(TypeError, match="Objective polynomial's and constraints' variable types do not match"):
        constrained_objective.get_all_variables()

    constrained_objective["wrong"] = ConstraintBinary(PolyBinary({((X, "a"),): 1, ((X, "b"),): 1}), 0, 1)
    with pytest.raises(ValueError, match="Sorting failed"):
        constrained_objective.get_all_variables()

    constrained_objective = ConstrainedObjective(PolyBinary({(4, 5): 1}),
                                                 {"constraint": ConstraintBinary(PolyBinary({(1, 2): 1}), 0, 0)})
    assert constrained_objective.get_all_variables() == [1, 2, 4, 5]

def test_get_redundant_constraints():
    """ test getting the name of the redundant constraints """
    objective_poly = PolyBinary({(0,): 7, (1, 0): 2, (): -1})
    quadratic_poly = PolyBinary({(0,): 7, (1, 0): 2, (): -3})
    quadratic_constraint = ConstraintBinary(quadratic_poly, 5, 6)
    constraints = {"quadratic": quadratic_constraint,
                   "again": ConstraintBinary(quadratic_poly, 5, 6),
                   "scaled": quadratic_constraint * 2,
                   "shifted": quadratic_constraint + 1,
                   "scaled_shifted": quadratic_constraint * 2 + 1}
    with pytest.warns(UserWarning, match=r"The ConstrainedObjective contains 4 redundant constraint\(s\)"):
        constrained_objective = ConstrainedObjective(objective_poly=objective_poly, constraints=constraints)
    exp_constraints = {"again", "scaled", "shifted", "scaled_shifted"}
    assert constrained_objective.get_redundant_constraints() == exp_constraints

    # the ordering is important: the first will be marked as non-redundant while the following are redundant
    constraints = {"scaled_shifted": quadratic_constraint * 2 + 1, "quadratic": quadratic_constraint}
    with pytest.warns(UserWarning, match=r"The ConstrainedObjective contains 1 redundant constraint\(s\)"):
        constrained_objective = ConstrainedObjective(objective_poly=objective_poly, constraints=constraints)
    exp_constraints = {"quadratic"}
    assert constrained_objective.get_redundant_constraints() == exp_constraints

def test_remove_redundant_constraints(constrained_objective):
    """ test redundant constraints methods """
    # Case 0: Removing constraints from a ConstraintObjective without redundant constraints
    assert constrained_objective.remove_redundant_constraints() == constrained_objective

    # Case 1: Linear constraints
    poly1 = PolyBinary({((X, 1, "red"),): 1, ((X, 1, "green"),): 1, ((X, 1, "blue"),): 1})
    poly2 = PolyBinary({((X, 2, "red"),): 2, ((X, 2, "green"),): 2, ((X, 2, "blue"),): 2})
    poly3 = PolyBinary({((X, 3, "red"),): 1, ((X, 3, "green"),): 1, ((X, 3, "blue"),): 1, (): 1})
    poly4 = PolyBinary({((X, 4, "red"),): 2, ((X, 4, "green"),): 2, ((X, 4, "blue"),): 2, (): 2})
    redundant_cons = {"again_1": ConstraintBinary(poly1, 1, 1),
                      "scaled_2": ConstraintBinary(poly2, 2, 2),
                      "shifted_3": ConstraintBinary(poly3, 2, 2),
                      "scaled_shifted_4": ConstraintBinary(poly4, 4, 4)}
    with pytest.warns(UserWarning, match=r"The ConstrainedObjective contains 4 redundant constraint\(s\)"):
        red_constrained_objective = ConstrainedObjective(objective_poly=constrained_objective.objective_poly,
                                                         constraints=constrained_objective|redundant_cons)
    assert red_constrained_objective.remove_redundant_constraints() == constrained_objective

    # Case 2: Quadratic constraints
    objective_poly = PolyBinary({(0,): 7, (1, 0): 2, (): -1})
    quadratic_poly = PolyBinary({(0,): 7, (1, 0): 2, (): -3})
    quadratic_constraint = ConstraintBinary(quadratic_poly, 5, 6)
    constraints = {"quadratic": quadratic_constraint}
    redundant_constraints = {"again": ConstraintBinary(quadratic_poly, 5, 6),
                             "scaled": quadratic_constraint * 2,
                             "shifted": quadratic_constraint + 1,
                             "scaled_shifted": quadratic_constraint * 2 + 1}

    exp_constrained_objective = ConstrainedObjective(objective_poly=objective_poly, constraints=constraints)
    with pytest.warns(UserWarning, match=r"The ConstrainedObjective contains 4 redundant constraint\(s\)"):
        red_constrained_objective = ConstrainedObjective(objective_poly=objective_poly,
                                                         constraints=constraints|redundant_constraints)
    assert red_constrained_objective.remove_redundant_constraints() == exp_constrained_objective

def test_get_broken_constraints(constrained_objective, solution):
    """ test broken constraints """
    assert constrained_objective.get_broken_constraints(solution) == []
    solution[(X, 2, "green")] = 1
    assert constrained_objective.get_broken_constraints(solution) == [ONE_COLOR_PER_NODE + "_2"]

def test_check_validity(constrained_objective, solution):
    """ test validity of solution """
    assert constrained_objective.check_validity(solution)

    solution[(X, 2, "green")] = 1
    assert not constrained_objective.check_validity(solution)

def test_get_original_problem_solution(constrained_objective, solution):
    """ test solution extraction """
    exp_original_solution = {1: 'blue', 2: 'red', 3: 'green', 4: 'blue', 5: 'red'}
    assert constrained_objective.get_original_problem_solution(solution) == exp_original_solution

def test_init__check_consistency(instance):
    """ test the errors """
    # pylint: disable=protected-access
    with pytest.warns(DeprecationWarning, match="This way of instantiation is deprecated"):
        ExampleConstrainedObjective(instance=instance)
    with pytest.raises(NotImplementedError, match="Provide 'objective_poly' on initialization or "):
        ConstrainedObjective._get_objective_poly(None)
    with pytest.raises(NotImplementedError, match="Provide 'constraints' on initialization or "):
        ConstrainedObjective._get_constraints(None)
    with pytest.raises(ValueError, match="Currently only binary objective functions are supported for the"):
        ConstrainedObjective(PolyBinary({"x0" : 1}).to_ising(), {})
    # Variable Types
    with pytest.raises(TypeError, match="Objective polynomial's and constraints' variable types do not match"):
        ConstrainedObjective(PolyBinary({(("x", 1),): 1}),
                             {"constraint": ConstraintBinary(PolyBinary({(0,): 1, (1,): 2}), 0, 1)})
    with pytest.raises(TypeError, match="Objective polynomial's and constraints' variable types do not match"):
        ConstrainedObjective(PolyBinary({(("x", 1),): 1}),
                             {"constraint": ConstraintBinary(PolyBinary({("x",): 1, ("y",): 2}), 0, 2)})
    with pytest.raises(TypeError, match="Objective polynomial's and constraints' variable types do not match"):
        ConstrainedObjective(PolyBinary({(0, 1): 1}),
                             {"constraint": ConstraintBinary(PolyBinary({("x",): 1, ("y",): 2}), 0, 2)})
    with pytest.raises(ValueError, match="Sorting failed"):
        ConstrainedObjective(PolyBinary({(("x", 0), ("x", 1)): 1}),
                             {"constraint": ConstraintBinary(PolyBinary({(("x", "0"),): 1, (("x", "1"),): 2}), 0, 2)})

def test_is_compact(constrained_objective):
    """ test check for compact constrained objectives """
    assert not constrained_objective.is_compact()

    compact, _ = constrained_objective.compact()
    assert compact.is_compact()

    objective_poly = PolyBinary({(1, 2, 3): 1, (1,) : -1})
    constraint = ConstraintBinary(PolyBinary({(1, 2): 1, (2,): -1}), 0, 1)
    constrained_objective = ConstrainedObjective(objective_poly, {"constraint": constraint})
    assert not constrained_objective.is_compact()

    constrained_objective["additional_constraint"] = ConstraintBinary(PolyBinary({(0,): 1, (1,): 1}), 0, 1)
    assert constrained_objective.is_compact()

def test_compact(constrained_objective):
    """ test making the polynomials compact """
    exp_compact_poly = PolyBinary({(0, 3): 1, (0, 6): 1, (0, 12): 1, (1, 4): 1, (1, 7): 1, (1, 13): 1, (2, 5): 1,
                                   (2, 8): 1, (2, 14): 1, (3, 6): 1, (4, 7): 1, (5, 8): 1, (6, 9): 1, (6, 12): 1,
                                  (7, 10): 1, (7, 13): 1, (8, 11): 1, (8, 14): 1, (9, 12): 1, (10, 13): 1, (11, 14): 1})
    exp_compact_constraints = {'one_color_per_node_1': ConstraintBinary(PolyBinary({(0,): 1, (1,): 1, (2,): 1}), 1, 1),
                               'one_color_per_node_2': ConstraintBinary(PolyBinary({(3,): 1, (4,): 1, (5,): 1}), 1, 1),
                               'one_color_per_node_3': ConstraintBinary(PolyBinary({(6,): 1, (7,): 1, (8,): 1}), 1, 1),
                               'one_color_per_node_4': ConstraintBinary(PolyBinary({(9,): 1, (10,): 1, (11,): 1}), 1,1),
                               'one_color_per_node_5': ConstraintBinary(PolyBinary({(12,): 1, (13,): 1, (14,): 1}),1,1)}
    exp_var_mapping = VariableMapping({0: ('x', 1, 'blue'), 1: ('x', 1, 'green'), 2: ('x', 1, 'red'),
                                       3: ('x', 2, 'blue'), 4: ('x', 2, 'green'), 5: ('x', 2, 'red'),
                                       6: ('x', 3, 'blue'), 7: ('x', 3, 'green'), 8: ('x', 3, 'red'),
                                       9: ('x', 4, 'blue'), 10: ('x', 4, 'green'), 11: ('x', 4, 'red'),
                                       12: ('x', 5, 'blue'), 13: ('x', 5, 'green'), 14: ('x', 5, 'red')})
    compact, var_mapping = constrained_objective.compact()
    assert not compact.name
    assert compact.objective_poly == exp_compact_poly
    assert dict(compact) == exp_compact_constraints
    assert var_mapping == exp_var_mapping

    # nothing happens when applying again
    assert compact.compact() == (compact, None)

    # more variables in the constraints than in the objective_poly
    objective_poly = PolyBinary({(('x', 3, "blue"), ('x', 4, "blue")): 1, (('x', 3, "red"), ('x', 4, "red")): 1})
    constrained_objective = ConstrainedObjective(objective_poly, constraints=dict(constrained_objective))
    exp_compact_poly = PolyBinary({(6, 9): 1, (8, 11): 1})
    compact, var_mapping = constrained_objective.compact("new_name")
    assert compact.name == "new_name"
    assert compact.objective_poly == exp_compact_poly
    assert dict(compact) == exp_compact_constraints
    assert var_mapping == exp_var_mapping

def test_get_objective_terms__two_paths(constrained_objective, objective_terms):
    """ test objective terms creation by comparing it to constrained_objective """
    objective_terms_from_co = constrained_objective.get_objective_terms(COLORED_EDGES, combine_prefixes=())
    one_color_per_node_polys = dict(itertools.islice(objective_terms_from_co.items(), 5))
    one_color_per_node_sum_poly = sum(one_color_per_node_polys.values())
    assert objective_terms_from_co[COLORED_EDGES] == objective_terms[COLORED_EDGES]
    assert one_color_per_node_sum_poly == objective_terms[ONE_COLOR_PER_NODE]

    ots_from_co_combined = constrained_objective.get_objective_terms(COLORED_EDGES, combine_prefixes=ONE_COLOR_PER_NODE)
    assert ots_from_co_combined != objective_terms_from_co
    assert ots_from_co_combined == objective_terms

def test_constrained_objective_creation_no_impl():
    """ test constrained objective creation with constraint initialization """
    objective_poly = PolyBinary.read_from_string(OBJECTIVE_POLY_FLAT_STR)
    constraint = ConstraintBinary(PolyBinary.read_from_string("x0 + x1 + x2"), 1, 1)
    constraints = {ONE_COLOR_PER_NODE : constraint}
    constrained_objective = ConstrainedObjective(objective_poly, constraints)
    assert constrained_objective.objective_poly == objective_poly
    assert dict(constrained_objective) == constraints
    assert constrained_objective != ConstrainedObjective(PolyBinary({(0,): 1}), constraints)
    assert constrained_objective != constraints

    with pytest.warns(UserWarning, match="There is nothing in this ConstrainedObjective"):
        ConstrainedObjective(PolyBinary())

def test_reduce():
    """ test the reduction of the constrained objective """
    objective_poly = PolyBinary({((X, 1, 1), (Y, 0)): 1, ((X, 1, 1), (Y, 1)): 3, ((X, 1, 2), (Y, 2),): 2})
    constraints = {"max_one_x": ConstraintBinary(PolyBinary({((X, 1, 1),): 1, ((X, 1, 2),): 1}), 0, 1),
                   "exactly_one_y": ConstraintBinary(PolyBinary({((Y, 0),): 1, ((Y, 1),): 1, ((Y, 2),): 1}), 1, 1)}
    constrained_objective = ConstrainedObjective(objective_poly, constraints)

    # nothing happens
    reduced_obj_poly, reduced_constraints, reductions = constrained_objective.get_reductions()
    assert reduced_obj_poly == objective_poly
    assert reduced_constraints == constraints
    assert reductions == []

    # apply prepared reductions
    exp_objective_poly = PolyBinary({((R, X, 1, 1, Y, 0),): 1,
                                     ((R, X, 1, 1, Y, 1),): 3,
                                     ((R, X, 1, 2, Y, 2),): 2})
    reduced_constrained_objective = constrained_objective.reduce(use=[((R, X, 1, 1, Y, 0), (X, 1, 1), (Y, 0)),
                                                                      ((R, X, 1, 1, Y, 1), (X, 1, 1), (Y, 1)),
                                                                      ((R, X, 1, 2, Y, 2), (X, 1, 2), (Y, 2))])
    assert reduced_constrained_objective.objective_poly == exp_objective_poly
    assert dict(reduced_constrained_objective) == constraints

    # force the reduction
    exp_constraint_x_reduced = ConstraintBinary(PolyBinary({((X, 1, 1),): 1, ((X, 1, 2),): 1}), 0, 1)
    exp_constraint_y_reduced = ConstraintBinary(PolyBinary({((Y, 0),): 1, ((Y, 1),): 1, ((Y, 2),): 1}), 1, 1)
    exp_constraints = {'max_one_x': exp_constraint_x_reduced,
                       'exactly_one_y': exp_constraint_y_reduced}

    reduced_obj_poly, reduced_constraints, reductions = constrained_objective.get_reductions(force=True)
    assert reduced_obj_poly == exp_objective_poly
    assert reduced_constraints == exp_constraints
    assert reductions == [((R, X, 1, 1, Y, 0), (X, 1, 1), (Y, 0)),
                          ((R, X, 1, 1, Y, 1), (X, 1, 1), (Y, 1)),
                          ((R, X, 1, 2, Y, 2), (X, 1, 2), (Y, 2))]

    # get all reduction constraints
    exp_poly_dicts = [{(): 1, ((R, X, 1, 1, Y, 0),): 1, ((X, 1, 1),): -1},
                      {(): 1, ((R, X, 1, 1, Y, 0),): 1, ((Y, 0),): -1},
                      {(): 1, ((R, X, 1, 1, Y, 0),): 1, ((X, 1, 1),): -1, ((Y, 0),): -1},
                      {(): 1, ((R, X, 1, 1, Y, 1),): 1, ((X, 1, 1),): -1},
                      {(): 1, ((R, X, 1, 1, Y, 1),): 1, ((Y, 1),): -1},
                      {(): 1, ((R, X, 1, 1, Y, 1),): 1, ((X, 1, 1),): -1, ((Y, 1),): -1},
                      {(): 1, ((R, X, 1, 2, Y, 2),): 1, ((X, 1, 2),): -1},
                      {(): 1, ((R, X, 1, 2, Y, 2),): 1, ((Y, 2),): -1},
                      {(): 1, ((R, X, 1, 2, Y, 2),): 1, ((X, 1, 2),): -1, ((Y, 2),): -1}]
    exp_names = ['reduction_x_1_1_y_0_0', 'reduction_x_1_1_y_0_1', 'reduction_x_1_1_y_0_2', 'reduction_x_1_1_y_1_0',
                 'reduction_x_1_1_y_1_1', 'reduction_x_1_1_y_1_2', 'reduction_x_1_2_y_2_0', 'reduction_x_1_2_y_2_1',
                 'reduction_x_1_2_y_2_2']
    exp_constraints.update(dict(zip(exp_names, [ConstraintBinary(PolyBinary(d), 0, 1) for d in exp_poly_dicts])))
    exp_constrained_objective = ConstrainedObjective(exp_objective_poly, exp_constraints)
    reduced_constrained_objective = constrained_objective.reduce(force=True)
    assert reduced_constrained_objective == exp_constrained_objective

def test_reduce__two_steps():
    """ test reduction in two steps """
    # first step
    objective_poly = PolyBinary({((X, 1), (X, 2), (X, 3)): 1})
    constraint_x = ConstraintBinary(PolyBinary({((X, 2),): 1, ((X, 3),): 1}), 0, 1)
    constraints = {"max_one_x": constraint_x}
    constrained_objective = ConstrainedObjective(objective_poly, constraints)

    exp_objective_poly = PolyBinary({((R, X, 1, X, 2), (X, 3)): 1})
    exp_constraint_red = ConstraintBinary(PolyBinary({((R, X, 1, X, 2),): 1, ((X, 1), (X, 2)): -1}), 0, 0)
    exp_constraints = {"max_one_x": constraint_x, "reduction_x_1_x_2": exp_constraint_red}
    exp_constrained_objective = ConstrainedObjective(exp_objective_poly, exp_constraints)

    reduced_constrained_objective = constrained_objective.reduce(max_degree=2)
    assert reduced_constrained_objective == exp_constrained_objective

    # reduce further, without force
    exp_objective_poly = PolyBinary({((R, X, 1, X, 2), (X, 3)): 1})
    exp_constraints = {"max_one_x": constraint_x}
    exp_constraints.update(get_reduction_constraints((R, X, 1, X, 2), (X, 1), (X, 2)))
    exp_constrained_objective = ConstrainedObjective(exp_objective_poly, exp_constraints)

    reduced_constrained_objective2 = reduced_constrained_objective.reduce(1)
    assert reduced_constrained_objective2 == exp_constrained_objective

    # reduce further, with force
    exp_objective_poly = PolyBinary({((R, R, X, 1, X, 2, X, 3),): 1})
    exp_constraints = {"max_one_x": ConstraintBinary(PolyBinary({((X, 2),): 1, ((X, 3),): 1}), 0, 1)}
    exp_constraints.update(get_reduction_constraints((R, X, 1, X, 2), (X, 1), (X, 2)))
    exp_constraints.update(get_reduction_constraints((R, R, X, 1, X, 2, X, 3), (R, X, 1, X, 2), (X, 3)))
    exp_constrained_objective = ConstrainedObjective(exp_objective_poly, exp_constraints)

    reduced_constrained_objective3 = reduced_constrained_objective2.reduce(1, force=True)
    assert reduced_constrained_objective3 == exp_constrained_objective

    reduced_constrained_objective2_direct_force = reduced_constrained_objective.reduce(1, force=True)
    assert reduced_constrained_objective2_direct_force == exp_constrained_objective

def test_reduce__big_example(constrained_objective, exp_constrained_objective, instance):
    """ test the automatic reduction of the objective function and the constraint polynomials """
    # nothing happens
    reduced_constrained_objective = constrained_objective.reduce()
    assert reduced_constrained_objective == exp_constrained_objective

    # several reductions are applied
    exp_constraints = dict(exp_constrained_objective)
    exp_objective_poly = PolyBinary({((R, X, n1, c, X, n2, c),): 1
                                     for n1, n2 in instance.edges for c in instance.colors})
    exp_reductions = {((R, X, n1, c, X, n2, c), (X, n1, c), (X, n2, c))
                      for n1, n2 in instance.edges for c in instance.colors}
    reduced_obj_poly, reduced_constraints, reductions = constrained_objective.get_reductions(force=True)
    assert reduced_obj_poly == exp_objective_poly
    assert reduced_constraints == exp_constraints
    assert set(reductions) == exp_reductions

    for n1, n2 in instance.edges:
        for c in instance.colors:
            names = [f"reduction_x_{n1}_{c}_x_{n2}_{c}_0",
                     f"reduction_x_{n1}_{c}_x_{n2}_{c}_1",
                     f"reduction_x_{n1}_{c}_x_{n2}_{c}_2"]
            polys = [PolyBinary({((R, X, n1, c, X, n2, c),): 1, ((X, n1, c),): -1, (): 1}),
                     PolyBinary({((R, X, n1, c, X, n2, c),): 1, ((X, n2, c),): -1, (): 1}),
                     PolyBinary({((R, X, n1, c, X, n2, c),): 1, ((X, n1, c),): -1, ((X, n2, c),): -1, (): 1})]
            exp_constraints.update(dict(zip(names, [ConstraintBinary(p, 0, 1) for p in polys])))
    exp_constrained_objective = ExampleConstrainedObjective(exp_objective_poly, exp_constraints)
    reduced_constrained_objective = constrained_objective.reduce(force=True)
    assert reduced_constrained_objective == exp_constrained_objective

def test_get_objective_terms(constrained_objective, exp_objective_terms):
    """ test objective terms creation """
    objective_terms = constrained_objective.get_objective_terms(COLORED_EDGES)
    assert objective_terms == exp_objective_terms

    # without explicit objective polynomial
    constrained_objective.objective_poly = PolyBinary()
    objective_terms = constrained_objective.get_objective_terms(COLORED_EDGES)
    assert not COLORED_EDGES in objective_terms

def test_get_objective_terms__reduction():
    """ test objective terms creation with automatic reduction """
    objective_poly = PolyBinary({(("x", 1), ("x", 2), ("x", 3)): 1, (("x", 1),) : -1})
    constraint = ConstraintBinary(PolyBinary({(("x", 1), ("x", 2)): 1, (("x", 2),): -1}), 0, 1)
    constrained_objective = ConstrainedObjective(objective_poly, {"constraint": constraint})

    # without checking for special constraints everything is reduced
    objective_terms = constrained_objective.get_objective_terms(check_special_constraints=False)
    exp_objective_poly = PolyBinary({((X, 1),): -1, ((R, X, 1, X, 2), (X, 3)): 1})
    exp_reduction_poly = PolyBinary({((R, X, 1, X, 2),): 3,
                                     ((R, X, 1, X, 2), (X, 1)): -2,
                                     ((R, X, 1, X, 2), (X, 2)): -2,
                                     ((X, 1), (X, 2)): 1})
    exp_constraint_poly = PolyBinary({(): 0,
                                      (('constraint_slack', 0),): 1,
                                      (('constraint_slack', 0), (R, X, 1, X, 2)): -2,
                                      (('constraint_slack', 0), (X, 2)): 2,
                                      ((R, X, 1, X, 2),): 1,
                                      ((R, X, 1, X, 2), (X, 2)): -2,
                                      ((X, 2),): 1})
    assert len(objective_terms) == 3
    assert objective_terms["objective"] == exp_objective_poly
    assert objective_terms["reduction"] == exp_reduction_poly
    assert objective_terms["constraint"] == exp_constraint_poly

    # with checking no slack variables are introduced
    objective_terms = constrained_objective.get_objective_terms()
    exp_constraint_poly = PolyBinary({((X, 2),): 1, ((R, X, 1, X, 2), (X, 2)): -1})
    assert len(objective_terms) == 3
    assert objective_terms["objective"] == exp_objective_poly
    assert objective_terms["reduction"] == exp_reduction_poly
    assert objective_terms["constraint"] == exp_constraint_poly

    # even if we reduce beforehand, the objective terms remain the same
    constrained_objective_reduced = constrained_objective.reduce()
    assert len(constrained_objective_reduced) == 4  # 1 original constraint and 3 reductions constraints
    objective_terms_reduced = constrained_objective_reduced.get_objective_terms()
    assert objective_terms_reduced == objective_terms
    # TODO: add more tests here!
